import pickle
from spectramanipulator.settings import Settings
from spectramanipulator.logger import Logger
import spectramanipulator
import sys

COMPRESS_LEVEL = 3

# used for backward compatibility
class SafeUnpickler(pickle.Unpickler):
    # help from https://www.programcreek.com/python/example/1606/pickle.Unpickler
    # https://stackoverflow.com/questions/13398462/unpickling-python-objects-with-a-changed-module-path

    def find_class(self, module, name):
        """
        Overridden from the original 'Unpickler' class. Needed to rebuild PyMod object which have
        complex modules names. 'Unpickler' rebuilds objects using the 'fully qualified' name
        reference of their classes (the class name is pickled, along with the name of the module the
        class is defined in). Since PyMOL plugin modules may yield different 'fully qualified' names
        depending on the system, PyMod objects are rebuilt using only the name of their classes.
        """

        if name == 'SpectrumList':  # for very old backward compatibility
            return list

        # Try the standard routine of pickle.
        try:
            # print(module)
            __import__(module)
        except ImportError:  # for backward compatibility when the program was not a python Package
            print(f'import of {module} failed, adding spectramanipulator.')
            module = f'spectramanipulator.{module}'
            __import__(module)

        mod = sys.modules[module]
        klass = getattr(mod, name)
        return klass


class Project:

    def __init__(self, generic_item=None, *args):

        # self.spectra_list = spectra_list  # used in older program versions
        self.generic_item = generic_item
        self.settings = Settings()  # saves the static attributes to instance attributes that are saved

        self.args = args
        self.__version__ = spectramanipulator.__version__

    def serialize(self, filepath):
        try:
            with open(filepath, 'bw') as file:
                pickle.dump(self, file)

            # try:
            # except:
            #     pass

            # with bz2.BZ2File(filepath, 'w', compresslevel=COMPRESS_LEVEL) as file:
            #     pickle.dump(self, file)

        # except pickle.PicklingError as err:
        #     Logger.console_message("Unable to save current project:\n{}\n{}".format(err.__str__(), err.__traceback__))
        #     raise Exception
        except Exception as err:
            Logger.message("Unable to save current project:\n{}".format(err.__str__()))
            raise

    @staticmethod
    def deserialize(filepath):

        # instance = pickle.load(open(filepath, 'rb'))

        instance = SafeUnpickler(open(filepath, 'rb')).load()
        return instance

        # try:
        #     try:
        #         # with open(filepath, 'br') as file:
        #         #     instance = pickle.load(file)
        #         instance = SafeUnpickler(open(filepath, 'rb')).load()
        #     except:  # for maintaining compatibility
        #         import bz2
        #         with bz2.BZ2File(filepath, 'r', compresslevel=COMPRESS_LEVEL) as file:
        #             instance = pickle.load(file)
        #
        #     return instance
        # # except pickle.UnpicklingError as err:
        # #     Logger.send_message("Unable to load {}:\n{}\n{}".format(filepath, err.__str__(), err.__traceback__))
        # #     raise
        # except Exception as err:
        #     Logger.message("Unable to load {}.\n{}".format(filepath, err.__str__()))









from PyQt5 import QtCore, QtGui, QtWidgets

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import *
from ..singleton import Singleton
import sys


class GenericInputDialog(Singleton, QtWidgets.QDialog):

    # static variables
    is_opened = False

    def __init__(self, widget_list=None, label_text='Some descriptive text...', title='GenericInputDialog',
                 parent=None, set_result=None, flags=Qt.WindowStaysOnTopHint | Qt.MSWindowsFixedSizeDialogHint):

        super(GenericInputDialog, self).__init__(parent, flags)

        # # disable resizing of the window,
        # # help from https://stackoverflow.com/questions/16673074/in-qt-c-how-can-i-fully-disable-resizing-a-window-including-the-resize-icon-w
        # self.setWindowFlags(Qt.Dialog | Qt.MSWindowsFixedSizeDialogHint)

        self.set_result = set_result

        self.setWindowTitle(title)
        # self.accepted = False

        self.button_box = QtWidgets.QDialogButtonBox(self)
        self.button_box.setOrientation(QtCore.Qt.Horizontal)
        self.button_box.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel | QtWidgets.QDialogButtonBox.Ok)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        self.VLayout = QVBoxLayout()
        # self.VLayout.setContentsMargins(0, 0, 0, 0)

        self.label = QLabel(label_text)

        self.VLayout.addWidget(self.label)

        self.grid_layout = QtWidgets.QGridLayout()
        self.grid_layout.setSpacing(7)
        self.grid_layout.setContentsMargins(15, 0, 0, 0)

        if widget_list is None:
            widget_list = [("par1", QLineEdit("some  text")),
                           ("par2", QLineEdit("another  text")),
                           ("par3", QCheckBox("...."))]

        self.widget_list = []

        for i, (label, widget) in enumerate(widget_list):
            if isinstance(label, str):
                # self.label_list.append(QLabel(label))
                label = QLabel(label)
                label.setWordWrap(True)
            if widget is not None:
                self.widget_list.append(widget)
                col_span = 2 if label is None else 1
                self.grid_layout.addWidget(widget, i, 1, 1, col_span)

            if label is not None:
                col_span = 2 if widget is None else 1
                self.grid_layout.addWidget(label, i, 0, 1, col_span)

        self.VLayout.addLayout(self.grid_layout)
        self.VLayout.addWidget(self.button_box)

        self.setLayout(self.VLayout)

        self.widget_list[0].setFocus()
        if isinstance(self.widget_list[0], QLineEdit):
            self.widget_list[0].selectAll()

    def show(self):
        if self.is_opened:
            self.activateWindow()
            self.setFocus()
            return
        self.is_opened = True
        super(GenericInputDialog, self).show()

    def accept(self):
        self.set_result()
        # self.accepted = True
        self.is_opened = False
        self._instance = None
        super(GenericInputDialog, self).accept()

    def reject(self):
        self.is_opened = False
        self._instance = None
        super(GenericInputDialog, self).reject()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = GenericInputDialog()
    Dialog.show()
    sys.exit(app.exec_())


import numpy as np
import os

from copy import deepcopy
import math

from scipy.signal import savgol_filter
from scipy.interpolate import interp1d
from scipy.fftpack import fft, fftfreq
from scipy.integrate import cumtrapz

import csv
import io

# from numba import jit
# from typing import Union, List, Iterable


def group2mat(spectra):
    """Given a generator of spectra, it returns the x, 'name' values and matrix.
    The columns of data will remain columns are horizontaly stack together.

    Parameters
    ----------
    spectra : iterable
        Spectra.

    Returns
    -------
    out : tuple
        x_values, converted_names as y values, matrix with [x, y] dimension"""

    header_vals_temp = []

    x = None
    matrix = None
    err = False

    for sp in spectra:
        try:
            header_vals_temp.append(float(sp.name.replace(',', '.').strip()))
        except ValueError:
            err = True

        if x is None:
            x = sp.data[:, 0]
        else:
            if x.shape[0] != sp.data.shape[0]:
                raise ValueError("Spectra do not have the same dimension within the group. Unable to perform the operation.")
            if not np.allclose(x, sp.data[:, 0]):
                raise ValueError("Spectra do not have the same x values within the group. Unable to perform the operation.")

        matrix = sp.data[:, 1] if matrix is None else np.vstack((matrix, sp.data[:, 1]))

    y = None if err else np.asarray(header_vals_temp)

    return x, y, matrix.T


class Spectrum(object):

    def __init__(self, data, filepath=None, name='', group_name='', color=None, line_width=None, line_alpha=255,
                 line_type=None, symbol=None, symbol_brush=None, sym_brush_alpha=255, symbol_fill=None,
                 sym_fill_alpha=255, symbol_size=8, plot_legend=True):
        """
        Class that holds the spectrum object as a 2D array (dimensions n x 2) where n is a number
        of points in spectrum and includes various functions used for data manipulation with them.
        Spectrum is stored in variable *data* as a numpy ndarray.

        Parameters
        ----------
        data : {numpy.ndarray, None}
            This argument represents data, this must be 2D numpy array.
            First column must contain x values and second y values. Data are automatically sorted according to
            first column.
        filepath : str, optional
            The path to the file the spectrum was imported from.
        name : str, optional
            Name of the spectrum.
        group_name : str, optional
            Group name of the spectrum (if inside group).
        color : {str, tuple}, optional
            Color of the spectrum, use string for common colors (eg. 'black', 'blue', 'red', ...)
            or tuple - red, green, blue, alpha components from 0 - 255, eg. (255, 0, 0, 255).
            If None, user defined color scheme will be used.
        line_width : {int, float}, optional
            Sets the line width of the plotted line. If None, user defined color scheme will be used.
        line_type : int 0-6, optional
            Sets the line type of the plotted line. See https://doc.qt.io/archives/qt-4.8/qt.html#PenStyle-enum
            for line types. If None, user defined color scheme will be used.
        plot_legend : bool
            Sets, if legend will be plotted for this spectrum.

        Attributes
        ----------
        data : numpy.ndarray
            A raw data of spectrum object, dimensions are (n x 2) where n is a number of points in spectrum.
            First column represents *x* values, second column *y* values.
        name : str
            The name of the spectrum as is displayed in Tree Widget.
        filepath : {str, None}
            If the spectrum was imported from file, this variable stored the path to the file it was imported from.
        color : {str, tuple, None}
            Color of the spectrum.
        line_width : {int, float, None}
            Line width of the plotted line.
        line_type : {int 0-6, None}
            Line type.
        plot_legend : bool
            If True, name of this spectrum will be added to legend. If False, legend will not contain this spectrum.
        """
        # sort according to first column,numpy matrix, 1. column wavelength, 2. column absorbance
        self.data = np.asarray(data[data[:, 0].argsort()], dtype=np.float64) if data is not None else None
        self.filepath = filepath
        self.name = name
        self.group_name = '' if group_name is None else group_name

        self.color = color  # line color
        self.line_width = line_width
        self.line_type = line_type
        self.plot_legend = plot_legend

        self.symbol = symbol
        self.symbol_brush = symbol_brush
        self.symbol_fill = symbol_fill
        self.symbol_size = symbol_size

        self.line_alpha = line_alpha
        self.sym_brush_alpha = sym_brush_alpha
        self.sym_fill_alpha = sym_fill_alpha

    @classmethod
    def from_xy_values(cls, x_values, y_values, name='', group_name='', color=None, line_width=None, line_alpha=255,
                       line_type=None, symbol=None, symbol_brush=None, sym_brush_alpha=255, symbol_fill=None,
                       sym_fill_alpha=255, symbol_size=8, plot_legend=True):
        """
        Creates the Spectrum object from separate x and y data variables. The dimensions of x_values
        and y_values must be the same and contain numbers. The other parameters are the same
        as for __init__ method.

        Parameters
        ----------
        x_values : iterable, eg. list, tuple, ndarray
            Iterable that represents x values.
        y_values : iterable, eg. list, tuple, ndarray
            Iterable that represents y values.

        Returns
        -------
        out : :class:`Spectrum`
            Spectrum object.


        Raises
        ------
        ValueError
            If *x_values* or *y_values* have not the same dimension or do not contain numbers.
        """
        try:
            if len(x_values) != len(y_values):
                raise ValueError("Length of x_values and y_values must match.")

            x_data = np.asarray(x_values, dtype=np.float64)
            y_data = np.asarray(y_values, dtype=np.float64)
        except ValueError:
            raise
        except:
            raise ValueError(
                "Argument error, x_values and y_values must be iterables and contain numbers.")

        data = np.vstack((x_data, y_data)).T

        return cls(data, name=name, group_name=group_name, color=color, line_width=line_width, line_alpha=line_alpha,
                   line_type=line_type, symbol=symbol, symbol_brush=symbol_brush, sym_brush_alpha=sym_brush_alpha,
                   symbol_fill=symbol_fill, sym_fill_alpha=sym_fill_alpha, symbol_size=symbol_size,
                   plot_legend=plot_legend)

    def add_to_list(self):
        """Adds this spectrum to the Tree Widget. Only works in spectramanipulator's Console Widget."""
        try:
            from user_namespace import add_to_list
            add_to_list(self)
        except ImportError:
            pass

    def _redraw_all_spectra(self):
        pass

    def _update_view(self):
        pass

    def set_plot_legend(self, plot_legend=True, redraw_spectra=True):
        """
        Sets whether to plot legend for this spectrum or not.

        Parameters
        ----------
        plot_legend : bool
            Default True.
        redraw_spectra : bool
            If True (default), spectra will be redrawn.
        """
        self.plot_legend = plot_legend
        if redraw_spectra:
            self._redraw_all_spectra()

    def set_style(self, color=None, line_width=None, line_type=None, redraw_spectra=True):
        """
        Set color, line width and line type of plotted spectrum.

        Parameters
        ----------
        color : {str, tuple}, optional
            Color of the spectrum, use string for common colors (eg. 'black', 'blue', 'red', ...)
            or tuple - red, green, blue, alpha components from 0 - 255, eg. (255, 0, 0, 255).
            If None (default), user defined color scheme will be used.
        line_width : {int, float}, optional
            Sets the line width of the plotted line. If None (default), user defined color scheme will be used.
        line_type : int 0-6, optional
            Sets the line type of the plotted line. See https://doc.qt.io/archives/qt-4.8/qt.html#PenStyle-enum
            for line types. If None (default), user defined color scheme will be used.
        redraw_spectra : bool
            If True (default), spectra will be redrawn.
        """
        # if color is not None:
        self.color = color
        # if line_width is not None:
        self.line_width = line_width
        # if line_type is not None:
        self.line_type = line_type
        if redraw_spectra:
            # from user_namespace import redraw_all_spectra
            self._redraw_all_spectra()

    def set_default_style(self, redraw_spectra=False):
        """
        Sets `color`, `line_width` and `line_type` to None. User defined color scheme will be used.

        Parameters
        ----------
        redraw_spectra : bool
            If True (default), spectra will be redrawn.
        """
        self.color = None
        self.line_width = None
        self.line_type = None

        if redraw_spectra:
            # from user_namespace import redraw_all_spectra
            self._redraw_all_spectra()

    @property
    def x(self):
        """Returns the x values of this spectrum.

        :return: ndarray
        """
        return self.data[:, 0]

    @x.setter
    def x(self, array):
        """Sets the x values of this spectrum.

        :return: ndarray
        """
        self.data[:, 0] = array

    @property
    def y(self):
        """Returns the y values of this spectrum.

        :return: ndarray
        """
        return self.data[:, 1]

    @y.setter
    def y(self, array):
        """Sets the y values of this spectrum.

        :return: ndarray
        """
        self.data[:, 1] = array

    def length(self):
        """Returns the number of point that this spectrum has.

        :return: float
        """
        return self.data.shape[0]

    def x_min(self):
        """Returns the the first x value.

        :return: float
        """
        return self.data[0, 0]

    def x_max(self):
        """Returns the the last x value.

        :return: float
        """
        return self.data[-1, 0]

    def spacing(self):
        """
        Returns a type and spacing of the x values:
            * r - probably regular spacing for all points
            * i - probably irregular spacing for all points

        The r/i distinction is made by checking differences between first and the end x values.
        Value is computed as average spacing:  x_max - x_min / number of points

        :return: str
        """

        spacing = (self.data[-1, 0] - self.data[0, 0]) / (self.data.shape[0] - 1)  # average spacing
        s_type = 'r'  # regular spacing
        if self.data.shape[0] > 2:
            # irregular spacing
            x_diff = self.data[1:, 0] - self.data[:-1, 0]  # x differences
            if not np.allclose(x_diff, x_diff[0]):
                s_type = 'i'

        return s_type + "{:.3g}".format(spacing)

    # https://stackoverflow.com/questions/15382076/plotting-power-spectrum-in-python
    def power_spectrum(self, redraw_spectra=True):
        """Calculates the power spectrum using FFT and replaces the data values. Power spectrum is normalized to 1.

        Parameters
        ----------
        redraw_spectra : bool
            If True (default), spectra will be redrawn (and Tree View updated).
        """

        n = self.data.shape[0]  # number of points in current spectrum

        vals = np.abs(fft(self.data[:, 1])) ** 2
        vals /= np.max(vals)
        freq = fftfreq(n, (self.data[-1, 0] - self.data[0, 0]) / (n - 1))

        filter = freq >= 0  # pick only frequencies >= 0

        output = np.vstack((freq[filter], vals[filter])).T

        self.data = output

        if redraw_spectra:
            self._redraw_all_spectra()
            self._update_view()

        return self

    def savgol(self, window_length, poly_order, redraw_spectra=True):
        """
        Applies Savitsky-Golay filter to a spectrum. Based on `scipy.signal.savgol_filter <https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.signal.savgol_filter.html>`_.

        Parameters
        ----------
        window_length : int
            Length of window that is used in the algorithm, must be positive odd integer.
        poly_order : int
            Polynomial order used in the algorithm, must be >= 1.
        redraw_spectra : bool
            If True (default), spectra will be redrawn (and Tree View updated).
        """
        window_length = int(window_length)
        poly_order = int(poly_order)

        if poly_order < 1:
            raise ValueError("Polynomial order must be > 0.")

        if poly_order >= window_length:
            raise ValueError("Polynomial order must be less than window_length.")

        # window_length must be odd number
        window_length += int(not window_length % 2)

        self.data[:, 1] = savgol_filter(self.data[:, 1], window_length, poly_order)

        if redraw_spectra:
            self._redraw_all_spectra()

        return self

    def gradient(self, edge_order=1, redraw_spectra=True):
        """
        Calculates a gradient using central differences and replaces the original values.

        Parameters
        ----------
        edge_order : int
            1 or 2, Gradient is calculated using N-th order accurate differences at the boundaries. Default: 1.

        redraw_spectra : bool
            If True (default), spectra will be redrawn (and Tree View updated).
        """

        self.data[:, 1] = np.gradient(self.data[:, 1], self.data[:, 0], edge_order=edge_order)

        if redraw_spectra:
            self._redraw_all_spectra()
            self._update_view()
        return self

    def differentiate(self, n=1, redraw_spectra=True):
        """
        Calculates a derivative spectrum and replaces the values. Length of the spectrum decreases by 1.

        Parameters
        ----------
        n : int
            The number of time s the values are differentiated. Default 1.
        redraw_spectra : bool
            If True (default), spectra will be redrawn (and Tree View updated).
        """

        new_data = np.zeros((self.data.shape[0] - n, 2))

        x_diffs = np.diff(self.data[:, 0], 1)[n - 1:]  # changes in x data

        new_data[:, 1] = np.diff(self.data[:, 1], n) / (x_diffs ** n)  # dy/dx
        new_data[:, 0] = self.data[n:, 0]

        self.data = new_data

        if redraw_spectra:
            self._redraw_all_spectra()
            self._update_view()
        return self

    def integrate(self, int_constant=0, redraw_spectra=True):
        """Integrates the spectrum using treapezoidal integration method, spectrum y values will
        be replaced by integrated values.

        Parameters
        ----------
        int_constant : {int, float}
            Integration constant, default 0.
        redraw_spectra : bool
            If True (default), spectra will be rdrawn.
        """
        self.data[:, 1] = cumtrapz(self.data[:, 1], self.data[:, 0], initial=int_constant)

        if redraw_spectra:
            self._redraw_all_spectra()

        return self

    def integral(self, x0=None, x1=None):
        """Calculate cumulative integral of spectrum at specified x range [x0, x1] by trapezoidal integration method.
        If x0, x1 are None, all spectrum will be integrated.

        Parameters
        ----------
        x0 : {int, float, None}
            First x value. If both x0 and x1 are None, all spectrum will be integrated.
        x1 : {int, float, None}
            Last x value. If both x0 and x1 are None, all spectrum will be integrated.

        Returns
        -------
        out : float
            Integral - scalar.
        """

        start_idx = 0
        end_idx = self.data.shape[0]  # not exactly end index, this is actually end_index + 1

        if x0 is not None or x1 is not None:
            if x0 >= x1:
                raise ValueError(f"Argument error, x0 ({x0}) cannot be larger or equal than x1 ({x1}).")

            start_idx = Spectrum.find_nearest_idx(self.data[:, 0], x0)
            end_idx = Spectrum.find_nearest_idx(self.data[:, 0], x1) + 1

        return np.trapz(self.data[start_idx:end_idx, 1], self.data[start_idx:end_idx, 0])

    def baseline_correct(self, x0, x1, redraw_spectra=True):
        """Subtracts the average of y data from the specified x range [x0, x1] from y values.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        redraw_spectra : bool
            If True (default), spectra will be redrawn.
        """

        if x0 > x1:
            raise ValueError(f"Argument error, x0 ({x0}) cannot be larger than x1 ({x1}).")

        # get nearby indexes of user defined x values
        start_idx = Spectrum.find_nearest_idx(self.data[:, 0], x0)
        end_idx = Spectrum.find_nearest_idx(self.data[:, 0], x1) + 1

        # calculate the average of y values over the selected range
        avrg = np.average(self.data[start_idx:end_idx, 1])

        # subtract the average from y values
        self.data[:, 1] -= avrg

        if redraw_spectra:
            self._redraw_all_spectra()

        return self

    def normalize(self, x0, x1, redraw_spectra=True):
        """Finds an y maximum at specified x range [x0, x1] and divide all y values by this maximum.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        redraw_spectra : bool
            If True (default), spectra will be redrawn.
        """

        if x0 > x1:
            x0, x1 = x1, x0

        # if x0 > x1:
        #     raise ValueError(f"Argument error, x0 ({x0}) cannot be larger than x1 ({x1}).")

        # get nearby indexes of user defined x values
        start_idx = Spectrum.find_nearest_idx(self.data[:, 0], x0)
        end_idx = Spectrum.find_nearest_idx(self.data[:, 0], x1) + 1

        y_max = np.max(self.data[start_idx:end_idx, 1])

        # normalize y values
        self.data[:, 1] /= y_max

        if redraw_spectra:
            self._redraw_all_spectra()

        return self

    def find_maximum(self, x0=None, x1=None):
        """Returns a point (x and y value), which belongs to a maximum y value in a specified x range [x0, x1].

        Parameters
        ----------
        x0 : {int, float, None}
            First x value. If both x0 and x1 are None, all spectrum will be searched.
        x1 : {int, float, None}
            Last x value. If both x0 and x1 are None, all spectrum will be searched.

        Returns
        -------
        out : tuple - (x, y)
            A highest point in spectrum.
        """
        start_idx = 0
        end_idx = self.data.shape[0]  # not exactly end index, this is actually end_index + 1

        if x0 is not None and x1 is not None:
            if x0 >= x1:
                raise ValueError(f"Argument error, x0 ({x0}) cannot be larger or equal than x1 ({x1}).")

            start_idx = Spectrum.find_nearest_idx(self.data[:, 0], x0)
            end_idx = Spectrum.find_nearest_idx(self.data[:, 0], x1) + 1

        max_idx = np.argmax(self.data[start_idx:end_idx, 1])

        # return the x a y max value
        return self.data[start_idx + max_idx, 0], self.data[start_idx + max_idx, 1]

    def interpolate(self, spacing=1, kind='linear', redraw_spectra=True):
        """Interpolates the spectrum, based on `scipy.interpolation.interp1d <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>`_.

        Parameters
        ----------
        spacing : {int, float}
            Sets the spacing between output x values, default 1.
        kind : str
            Sets the kind of interpolation method used, eg. 'linear', 'quadratic', etc. For more, refer to
            `interp1d documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>`_.
            Default 'linear'.
        redraw_spectra : bool
            If True (default), spectra will be redrawn (and Tree View updated).
        """

        data_min = self.data[0, 0]
        data_max = self.data[-1, 0]

        if spacing > data_max - data_min:
            raise ValueError(f"Spacing ({spacing}) cannot be larger that data itself. "
                             f"Resample method, spectrum {self.name}.")

        # new x data must lie inside a range of original x values
        x_min = spacing * int(np.ceil(data_min / spacing))
        x_max = spacing * int(np.floor(data_max / spacing))

        # length of array of new data
        n = int((x_max - x_min) / spacing + 1)

        new_data = np.zeros((n, 2))
        new_data[:, 0] = np.linspace(x_min, x_max, num=n)  # new x values

        # interp1d returns a function that takes x values as the argument and returns the interpolated y values
        f = interp1d(self.data[:, 0], self.data[:, 1], kind=kind, copy=False, assume_sorted=True)

        new_data[:, 1] = f(new_data[:, 0])  # interpolate the y values

        # change the reference of self.data to new_data
        self.data = new_data

        if redraw_spectra:
            self._redraw_all_spectra()
            self._update_view()

        return self

    def cut(self, x0, x1, redraw_spectra=True):
        """Cuts the spectrum to x range [x0, x1].

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        redraw_spectra : bool
            If True (default), spectra will be redrawn (and Tree View updated).
        """
        if x0 >= x1:
            raise ValueError(f"Argument error, x0 ({x0}) cannot be larger or equal than x1 ({x1}).")

        # get nearby indexes of user defined x values
        start_idx = Spectrum.find_nearest_idx(self.data[:, 0], x0)
        end_idx = Spectrum.find_nearest_idx(self.data[:, 0], x1) + 1

        if start_idx + 1 == end_idx:
            raise ValueError(
                "Oh... someone likes to cut a lot, but with these input parameters, resulting spectrum would have"
                " 1 point :). Unfortunately, cannot perform cut operation.")

        self.data = self.data[start_idx:end_idx, :]

        if redraw_spectra:
            self._redraw_all_spectra()
            self._update_view()

        return self

    def extend_by_zeros(self, x0, x1, redraw_spectra=True):
        """
        Extends this spectrum to a new x range [x0, x1] by zeros. The original data will
        be replaced. Spacing will be determined from current data.

        Parameters
        ----------
        x0 : {int, float}
            New fist x value.
        x1 : {int, float}
            New last x value.
        redraw_spectra : bool
            If True (default), spectra will be redrawn.
        """
        if x0 >= x1:
            raise ValueError(f"Argument error, x0 ({x0}) cannot be larger or equal than x1 ({x1}).")

        x_min = self.x_min()
        x_max = self.x_max()

        # get nearby indexes of user defined x values
        start_idx = Spectrum.find_nearest_idx(self.data[:, 0], x0)
        end_idx = Spectrum.find_nearest_idx(self.data[:, 0], x1) + 1

        if start_idx != 0 and end_idx != self.data.shape[0]:
            # nothing to extend
            return
            # raise ValueError("Nothing to extend.")

        x_dif = x_max - x_min
        spacing = x_dif / (self.data.shape[0] - 1)

        new_x_min = spacing * int(np.round(x0 / spacing, 0))

        min_stack = None
        try:
            num_min = int((x_min - new_x_min) / spacing + 1)
            min_lin_space = np.linspace(new_x_min, x_min, num=num_min)

            min_stack = np.zeros((2, num_min - 1))
            min_stack[0] = min_lin_space[:-1]
            min_stack = min_stack.T
        except ValueError:
            pass

        max_stack = None
        try:
            new_x_max = spacing * int(np.round(x1 / spacing, 0))

            num_max = int((new_x_max - x_max) / spacing + 1)
            max_lin_space = np.linspace(x_max, new_x_max, num=num_max)

            max_stack = np.zeros((2, num_max - 1))
            max_stack[0] = max_lin_space[1:]
            max_stack = max_stack.T
        except ValueError:
            pass

        if min_stack is not None and max_stack is not None:
            result = np.vstack((min_stack, self.data, max_stack))
        elif min_stack is not None:
            result = np.vstack((min_stack, self.data))
        else:
            result = np.vstack((self.data, max_stack))

        self.data = result

        if redraw_spectra:
            self._redraw_all_spectra()
            self._update_view()

        return self
    #
    # def _operation_check(self, other):
    #     """Checks if the shapes of self and other data array are the same and where
    #     first and last x value is the same. If not, raises ValueError."""
    #     if not isinstance(other, Spectrum):
    #         return
    #     if other.data.shape[0] != self.data.shape[0]:
    #         raise ValueError(
    #             f"Spectra '{self.name}' and '{other.name}' have not the same length (dimension). "
    #             "Unable to perform calculation.")
    #     if other.data[0, 0] != self.data[0, 0] or other.data[-1, 0] != self.data[-1, 0]:
    #         raise ValueError(
    #             f"Spectra '{self.name}' and '{other.name}' have different x ranges. "
    #             "Unable to perform calculation.")

    def _operation_check(self, other):
        """Checks if the shapes of self and other data array are the same and where
        first and last x value is the same. If not, raises ValueError."""
        if self.data.shape[0] != other.data.shape[0] or not np.allclose(self.data[:, 0], other.data[:, 0]):
            raise ValueError(
                f"Spectra '{self.name}' and '{other.name}' have not share the same x values "
                f"(or have different dimension). Unable to perform operation.")

    def _arithmetic_operation(self, other, func_operation):

        if not isinstance(other, (Spectrum, float, int)):
            raise ValueError("Cannot perform calculation of Spectrum with {}".format(type(other)))

        # another spectrum
        if isinstance(other, Spectrum):
            self._operation_check(other)
            y_data = func_operation(self.data[:, 1], other.data[:, 1])
            other_str = other.name

        else:  # number
            y_data = func_operation(self.data[:, 1], other)
            other_str = str(other)

        return Spectrum.from_xy_values(self.data[:, 0], np.nan_to_num(y_data)), other_str

    def __add__(self, other):
        if isinstance(other, SpectrumList):
            return other.__radd__(self)
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 + s2)
        sp.name = f"{self.name} + {other_str}"

        return sp

    def __sub__(self, other):
        if isinstance(other, SpectrumList):
            return other.__rsub__(self)
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 - s2)
        sp.name = f"{self.name} - {other_str}"

        return sp

    def __mul__(self, other):
        if isinstance(other, SpectrumList):
            return other.__rmul__(self)
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 * s2)
        sp.name = f"{self.name} * {other_str}"

        return sp

    def __truediv__(self, other):   # self / other
        if isinstance(other, SpectrumList):
            return other.__rtruediv__(self)
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 / s2)
        sp.name = f"{self.name} / {other_str}"

        return sp

    def __radd__(self, other):   # other + self
        if not isinstance(other, (int, float)):
            raise ValueError("Cannot perform calculation of Spectrum with {}".format(type(other)))
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 + s1)
        sp.name = f"{other_str} + {self.name}"

        return sp

    def __rsub__(self, other):   # other - self
        if not isinstance(other, (int, float)):
            raise ValueError("Cannot perform calculation of Spectrum with {}".format(type(other)))
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 - s1)
        sp.name = f"{other_str} - {self.name}"

        return sp

    def __rmul__(self, other):   # other * self
        if not isinstance(other, (int, float)):
            raise ValueError("Cannot perform calculation of Spectrum with {}".format(type(other)))
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 * s1)
        sp.name = f"{other_str} * {self.name}"

        return sp

    def __rtruediv__(self, other):  # other / self
        if not isinstance(other, (int, float)):
            raise ValueError("Cannot perform calculation of Spectrum with {}".format(type(other)))
        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 / s1)
        sp.name = f"{other_str} / {self.name}"

        return sp

    def __pow__(self, power, modulo=None):  # self ** power
        if isinstance(power, SpectrumList):
            return power.__rpow__(self)
        sp, other_str = self._arithmetic_operation(power, lambda s1, s2: s1 ** s2)
        sp.name = f"{self.name} ** {other_str}"

        return sp

    def __rpow__(self, other):  # other ** self
        if not isinstance(other, (Spectrum, float, int)):
            raise ValueError("Cannot perform calculation of Spectrum with {}".format(type(other)))

        sp, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 ** s1)
        sp.name = f"{other_str} ** {self.name}"

        return sp

    def __neg__(self):
        sp = self * -1
        sp.name = f'-{self.name}'
        return sp

    def __pos__(self):
        return self.__copy__()

    @staticmethod
    def float_to_string(float_number, decimal_sep='.'):
        return str(float_number).replace('.', decimal_sep)

    def __str__(self, separator='\t', decimal_sep='.', new_line='\n', include_header=True):
        buffer = "Wavelength" + separator + self.name + new_line if include_header else ""
        buffer += new_line.join(
            separator.join(Spectrum.float_to_string(num, decimal_sep) for num in row) for row in self.data)
        return buffer

    @staticmethod
    def _list_to_stream(list_of_spectra, include_group_name=True, include_header=True, delimiter='\t',
                        decimal_sep='.', save_to_file=False, dir_path=None, extension=None,
                        x_data_name='Wavelength / nm'):

        # this generator iterates the ndarray data and yield returns a list of formatted row for csv writer
        def iterate_data(iterable):
            for row in iterable:
                yield [str(num).replace('.', decimal_sep) for num in row]

        if not isinstance(list_of_spectra, list):
            raise ValueError("Argument 'list_of_spectra' must be type of list.")

        buffer = ""

        dialect = csv.excel
        dialect.delimiter = delimiter
        dialect.lineterminator = '\n'
        dialect.quoting = csv.QUOTE_MINIMAL

        for i, node in enumerate(list_of_spectra):

            if save_to_file:
                name = node.name if isinstance(node, Spectrum) else node[0].group_name
                filepath = os.path.join(dir_path,
                                        (name if name != '' else 'Untitled{}'.format(i)) + extension)
                w = csv.writer(open(filepath, 'w', encoding='utf-8'), dialect=dialect)
            else:
                stream = io.StringIO('')
                w = csv.writer(stream, dialect=dialect)

            # export as group
            if isinstance(node, list):
                if len(node) == 0:
                    continue

                if include_group_name:
                    w.writerow([node[0].group_name])

                # add header if it is user defined
                if include_header:
                    w.writerow([x_data_name] + [sp.name for sp in node])

                # add row of wavelength, then, we will transpose the matrix, otherwise, we would have to reshape
                matrix = node[0].data[:, 0]

                # add absorbance data to matrix from all exported spectra
                for sp in node:
                    if sp.length() != node[0].length():
                        raise ValueError(
                            f"Spectra '{node[0].name}' and '{sp.name}' in group '{node[0].group_name}'"
                            f" have not the same length (dimension). Unable to export.")
                    if sp.data[0, 0] != node[0].data[0, 0] or sp.data[-1, 0] != node[0].data[-1, 0]:
                        raise ValueError(
                            f"Spectra '{node[0].name}' and '{sp.name}' in group '{node[0].group_name}'"
                            f" have not the same length (dimension). Unable to export.")
                    matrix = np.vstack((matrix, sp.data[:, 1]))

                matrix = matrix.T  # transpose

                # write matrix
                w.writerows(iterate_data(matrix))

            # export as single spectrum file
            if isinstance(node, Spectrum):
                if include_header:
                    w.writerow([x_data_name, node.name])

                w.writerows(iterate_data(node.data))

            if not save_to_file:
                buffer += stream.getvalue() + '\n'
                stream.close()

        if buffer != "":
            # return ret_buffer and remove the 2 new line characters that are at the end
            return buffer[:-2]

    @staticmethod
    def list_to_files(list_of_spectra, dir_path, extension, include_group_name=True, include_header=True,
                      delimiter='\t',
                      decimal_sep='.', x_data_name='Wavelength / nm'):

        Spectrum._list_to_stream(list_of_spectra, include_group_name, include_header, delimiter,
                                 decimal_sep, True, dir_path, extension, x_data_name)

    @staticmethod
    def list_to_string(list_of_spectra, include_group_name=True, include_header=True, delimiter='\t',
                       decimal_sep='.', x_data_name='Wavelength / nm'):

        return Spectrum._list_to_stream(list_of_spectra, include_group_name, include_header, delimiter,
                                        decimal_sep, False, x_data_name=x_data_name)

    @staticmethod
    def find_nearest_idx(array, value):
        """
        Finds nearest index of `value` in `array`. If value >  max(array), the last index of array
        is returned, if value < min(array), 0 is returned. Array must be sorted. Also works for value
        to be array, then array of indexes is returned.

        Parameters
        ----------
        array : ndarray
            Array to be searched.
        value : {int, float}
            Value.

        Returns
        -------
        out : int, np.ndarray
            Found nearest index/es to value/s.
        """
        # idx = np.searchsorted(array, value, side="left")
        # if idx > 0 and (idx == len(array) or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
        #     return idx - 1
        # else:
        #     return idx
        if isinstance(value, (int, float)):
            value = np.asarray([value])
        else:
            value = np.asarray(value)

        result = np.empty_like(value, dtype=int)
        for i in range(value.shape[0]):
            idx = np.searchsorted(array, value[i], side="left")
            if idx > 0 and (
                    idx == len(array) or math.fabs(value[i] - array[idx - 1]) < math.fabs(value[i] - array[idx])):
                result[i] = idx - 1
            else:
                result[i] = idx
        return result if result.shape[0] > 1 else result[0]

    @staticmethod
    def find_nearest(array, value):
        """
        Returns closest value in in `array`. If value >  max(array), the last value of array
        is returned, if value < min(array), first value of array is returned. Array must be sorted.

        Parameters
        ----------
        array : ndarray
            Array to be searched.
        value : {int, float}
            Value.

        Returns
        -------
        out : type of values in `array`
            Found nearest value.
        """
        idx = Spectrum.find_nearest_idx(array, value)
        return array[idx]

    def __copy__(self):
        """Deep copy the current instance as Spectrum object."""
        sp = Spectrum(None, filepath=self.filepath, name=self.name, group_name=self.group_name, color=self.color,
                      line_width=self.line_width, line_alpha=self.line_alpha, line_type=self.line_type, symbol=self.symbol,
                      symbol_brush=self.symbol_brush, sym_brush_alpha=self.sym_brush_alpha, symbol_fill=self.symbol_fill,
                      sym_fill_alpha=self.sym_fill_alpha, symbol_size=self.symbol_size, plot_legend=self.plot_legend)

        sp.data = deepcopy(self.data)
        return sp


class SpectrumList(object):

    def __init__(self, children=None, name=''):
        """
        Class that holds a group of spectra as a list and enables simple arithmetic calculations among groups of spectra, number and
        group and spectrum and group. It acts as a list, so the individual spectra can be accessed by putting square brackets
        after the instance of an object and putting appropriate index in them. Also, spectrum objects can be iterated the same way as list.
        Also, it enables other calculations on the group of spectra, like baseline correction, cutting, basically all the
        operation provided in :class:`Spectrum` class plus additional operations that can
        be only used for groups, like transposition, etc. ....

        Parameters
        ----------
        children : {list, None}
            List of :class:`Spectrum`, default None.
        name : str, optional
            The group name.

        Attributes
        ----------
        children : list of :class:`Spectrum`
            List of members in this group.
        name : str
            The group name as appeared in Tree Widget.
        """

        self.children = [] if children is None else children
        self.name = name

    def __len__(self):
        return self.children.__len__()

    def add_to_list(self, spectra=None):
        """Adds this group to the Tree Widget. Only works in spectramanipulator's Console Widget.

        Parameters
        ----------
        spectra : {list, None}
            If None, current members will be added, otherwise, spectra will be imported.
        """
        try:
            from user_namespace import add_to_list
            add_to_list(self if spectra is None else spectra)
        except ImportError:
            pass

    def _redraw_all_spectra(self):
        pass

    def _update_view(self):
        pass

    def get_names(self):
        """Returns names of all Spectrum objects as a list.

        :returns: list
        """

        return [sp.name for sp in self]

    def set_names(self, names):
        """Sets the names of the group and updates the Tree Widget and redraws the spectra.

        Parameters
        ----------
        names : iterable, eg. list, tuple
            New names that will be replaced. The length of names can be different from number
            of spectra in group (using a zip function).
        """

        if not isinstance(self[0], Spectrum):
            raise ValueError("Objects in list have to be type of Spectrum.")

        for sp, new_name in zip(self, names):
            sp.name = str(new_name)

        # from user_namespace import update_view, redraw_all_spectra
        self._update_view()
        self._redraw_all_spectra()

    def set_plot_legend(self, plot_legend=True):
        """
        Sets whether to plot legend for this group or not and redraws all spectra.

        Parameters
        ----------
        plot_legend : bool
            Default True.
        """

        for sp in self:
            sp.set_plot_legend(plot_legend, False)

        self._redraw_all_spectra()

    def set_style(self, color=None, line_width=None, line_type=None, redraw_spectra=True):
        """
        Sets color, line width and line type of all group and redraws all spectra.

        Parameters
        ----------
        color : {str, tuple}, optional
            Color of the spectrum, use string for common colors (eg. 'black', 'blue', 'red', ...)
            or tuple - red, green, blue, alpha components from 0 - 255, eg. (255, 0, 0, 255).
            If None (default), user defined color scheme will be used.
        line_width : {int, float}, optional
            Sets the line width of the plotted line. If None (default), user defined color scheme will be used.
        line_type : int 0-6, optional
            Sets the line type of the plotted line. See https://doc.qt.io/archives/qt-4.8/qt.html#PenStyle-enum
            for line types. If None (default), user defined color scheme will be used.
        """

        for sp in self:
            sp.set_style(color, line_width, line_type, False)
        if redraw_spectra:
            self._redraw_all_spectra()

    # def _iterate_items(self):
    #     for node in self:
    #         if isinstance(node, SpectrumList):
    #             for sp in node:
    #                 if not isinstance(sp, Spectrum):
    #                     raise ValueError("Objects in list have to be type of Spectrum.")
    #                 yield sp
    #             continue
    #         if isinstance(node, Spectrum):
    #             yield node
    #             continue
            raise ValueError("Objects in list have to be type of Spectrum.")

    def gradient(self, edge_order=1):
        """
        Calculates a gradients using central differences and replaces the original values.

        Parameters
        ----------
        edge_order : int
            1 or 2, Gradient is calculated using N-th order accurate differences at the boundaries. Default: 1.
        """
        for sp in self:
            sp.gradient(edge_order, False)

        self._redraw_all_spectra()
        self._update_view()

        return self

    def differentiate(self, n=1):
        """
        Calculates a derivatives for this group, replaces the values and redraws the spectra.
        Length of the spectrum decreases by 1.
        """
        for sp in self:
            sp.differentiate(n, False)

        self._redraw_all_spectra()
        self._update_view()

        return self

    def integrate(self, int_constant=0):
        """Integrates the group of spectra using trapezoidal integration method, spectra's y values will
        be replaced by integrated values and redraws the spectra.

        Parameters
        ----------
        int_constant : {int, float}
            Integration constant, default 0.
        """

        for sp in self:
            sp.integrate(int_constant, False)

        self._redraw_all_spectra()

        return self

    def integral(self, x0=None, x1=None):
        """Calculates cumulative integral for each spectrum in this group at specified x range [x0, x1] by trapezoidal integration method.
        If x0, x1 are None, all spectra will be integrated. Result is returned as ndarray.

        Parameters
        ----------
        x0 : {int, float, None}
            First x value. If both x0 and x1 are None, all spectra will be integrated.
        x1 : {int, float, None}
            Last x value. If both x0 and x1 are None, all spectra will be integrated.

        Returns
        -------
        out : numpy.ndarray
            Array of integrals.
        """

        results = []

        for sp in self:
            results.append(sp.integral(x0, x1))

        return np.asarray(results, dtype=np.float64)

    def find_maximum(self, x0=None, x1=None):
        """Returns an array of points (x and y value), which belongs to a maximum y value in a specified x range [x0, x1].
        First column contains x values, the second y values.

        Parameters
        ----------
        x0 : {int, float, None}
            First x value. If both x0 and x1 are None, all spectra will be searched.
        x1 : {int, float, None}
            Last x value. If both x0 and x1 are None, all spectra will be searched.

        Returns
        -------
        out : ndarray
            A highest points in spectra.
        """
        results = []

        for sp in self:
            results.append(sp.find_maximum(x0, x1))

        return np.asarray(results, dtype=np.float64)

    def get_y_values_at_x(self, x):
        """Returns y values at particular x value as an ndarray in this group of spectra.

        Parameters
        ----------
        x : {int, float}
            The x value.

        Returns
        -------
        out : ndarray
            Array of y values at this x value.
        """

        ret_list = []

        for sp in self:
            idx = Spectrum.find_nearest_idx(sp.data[:, 0], x)
            ret_list.append(sp.data[idx, 1])

        return np.asarray(ret_list, dtype=np.float64)

    def transpose(self, max_items=1000):
        """
        Transposes group, names of spectra in group will be taken as x values for transposed data, so these values
        must be convertible to int or float numbers. No text is allowed in these cells, only values.
        All x values of spectra in the group will become names in new group. The transposed group is added to Tree Widget.
        The operation is basically the same as copying the group of spectra to Excel, performs transposition of the matrix
        and copying back to spectramanipulator.

        Parameters
        ----------
        max_items : int
            Maximum number of items that will be added. If the transposition produce more than *max_items*,
            ValueError will be raised. Default 1000.
        """

        if len(self) < 2:
            raise ValueError("At least 2 items have to be in the group in order to perform transposition.")

        if not isinstance(self[0], Spectrum):
            raise ValueError("Objects in list have to be type of Spectrum.")

        x, y, matrix = group2mat(self.__iter__())

        if y is None:
            raise ValueError("Names of spectra cannot be parsed to float.")

        n = x.shape[0]
        assert n == matrix.shape[0]

        group_name = "Transpose of {}".format(self.name)
        spectra = []

        if n > max_items:
            raise ValueError(
                "Number of transposed items ({}) exceeded the maximum number ({}). Cannot transpose spectra.".format(
                    n, max_items))

        for i in range(n):
            sp = Spectrum.from_xy_values(y, matrix[i], name=str(x[i]), group_name=group_name)
            spectra.append(sp)

        self.add_to_list([spectra])

    def power_spectrum(self):
        """Calculates the power spectrum using FFT and replaces the data values and redraws the spectra. Power spectrum is normalized to 1.
        """

        for sp in self:
            sp.power_spectrum(False)

        self._redraw_all_spectra()
        self._update_view()

        return self

    def savgol(self, window_length, poly_order):
        """
        Applies Savitysky-Golay filter to a group of spectra and redraws the spectra. Based on `scipy.signal.savgol_filter <https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.signal.savgol_filter.html>`_.

        Parameters
        ----------
        window_length : int
            Length of window that is used in the algorithm, must be positive odd integer.
        poly_order : int
            Polynomial order used in the algorithm, must be >= 1.
        """

        for sp in self:
            sp.savgol(window_length, poly_order, False)

        self._redraw_all_spectra()

        return self

    def baseline_correct(self, x0, x1):
        """Subtracts the average of y data from the specified x range [x0, x1] from y values for each spectrum in
        this group and redraws the spectra.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """
        for sp in self:
            sp.baseline_correct(x0, x1, False)

        self._redraw_all_spectra()

        return self

    def cut(self, x0, x1):
        """Cuts all spectra in this group to x range [x0, x1] and redraws the spectra.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """
        for sp in self:
            sp.cut(x0, x1, False)

        self._redraw_all_spectra()
        self._update_view()

        return self

    def interpolate(self, spacing=1, kind='linear'):
        """Interpolates the spectra in this group and redraws them, based on `scipy.interpolation.interp1d <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>`_.

        Parameters
        ----------
        spacing : {int, float}
            Sets the spacing between output x values, default 1.
        kind : str
            Sets the kind of interpolation method used, eg. 'linear', 'quadratic', etc. For more, refer to
            `interp1d documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>`_.
            Default 'linear'.
        """
        for sp in self:
            sp.interpolate(spacing, kind, False)

        self._redraw_all_spectra()
        self._update_view()

        return self

    def normalize(self, x0, x1):
        """Finds an y maximum at specified x range [x0, x1] and divide all y values by this maximum. This
        operation is perfomred for each spectrum in this group. The spectra are redrawn.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """
        for sp in self:
            sp.normalize(x0, x1, False)

        self._redraw_all_spectra()

        return self

    def extend_by_zeros(self, x0, x1):
        """
        Extends spectra in this group to a new x range [x0, x1] by zeros and redraws the spectra. The original data will
        be replaced. Spacing will be determined from data of current spectrum.

        Parameters
        ----------
        x0 : {int, float}
            New fist x value.
        x1 : {int, float}
            New last x value.
        """
        for sp in self:
            sp.extend_by_zeros(x0, x1, False)

        self._redraw_all_spectra()
        self._update_view()

        return self

    def _arithmetic_operation(self, other, func_operation):
        """
        Perform an defined operation on groups of spectra (type Spectrum). func_operation is a pointer to function.
        This function takes 2 arguments that must be type of Spectrum.
        """
        # operation with another list, group + group
        if isinstance(other, (SpectrumList, np.ndarray)):
            if len(self) != len(other):
                raise ValueError("Cannot perform an operation on groups which contains different number of items.")
            if len(self) == 0:
                return SpectrumList()
            # if not isinstance(self[0], Spectrum) or not isinstance(other[0], Spectrum):
            #     raise ValueError("Objects in list have to be type of Spectrum.")
            ret_list = SpectrumList()
            for i in range(len(self)):
                ret_list.children.append(func_operation(self[i], other[i]))
            return ret_list, other.name if isinstance(other, SpectrumList) else 'ndarray'

        # operation with single spectrum, group + spectrum or with number, eg. group - 1
        if isinstance(other, (Spectrum, float, int)):
            if len(self) == 0:
                return SpectrumList()

            if not isinstance(self[0], Spectrum):
                raise ValueError("Objects in list have to be type of Spectrum.")
            ret_list = SpectrumList()
            for sp in self:
                ret_list.children.append(func_operation(sp, other))
            return ret_list, str(other) if isinstance(other, (float, int)) else other.name

        raise ValueError("Cannot perform calculation of SpectrumList with {}".format(type(other)))

    def __add__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 + s2)
        list.name = "{} + {}".format(self.name, other_str)
        return list

    def __sub__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 - s2)
        list.name = "{} - {}".format(self.name, other_str)
        return list

    def __mul__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 * s2)
        list.name = "{} * {}".format(self.name, other_str)
        return list

    def __truediv__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s1 / s2)
        list.name = "{} / {}".format(self.name, other_str)
        return list

    def __radd__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 + s1)
        list.name = "{} + {}".format(other_str, self.name)
        return list

    def __rsub__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 - s1)
        list.name = "{} - {}".format(other_str, self.name)
        return list

    def __rmul__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 * s1)
        list.name = "{} * {}".format(other_str, self.name)
        return list

    def __rtruediv__(self, other):
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 / s1)
        list.name = "{} / {}".format(other_str, self.name)
        return list

    def __neg__(self):
        list = SpectrumList()
        list.name = f'-{self.name}'
        for sp in self:
            list.children.append(-sp)
        return list

    def __pos__(self):
        return self.__copy__()

    def __pow__(self, power, modulo=None):  # self ** power
        list, other_str = self._arithmetic_operation(power, lambda s1, s2: s1 ** s2)
        list.name = "{} ** {}".format(self.name, other_str)
        return list

    def __rpow__(self, other):  # other ** self
        list, other_str = self._arithmetic_operation(other, lambda s1, s2: s2 ** s1)
        list.name = "{} ** {}".format(other_str, self.name)
        return list

    def __getitem__(self, item):
        return self.children[item]

    def __iter__(self):
        return iter(self.children)

    def __copy__(self):
        """Deep copy this instance as SpectrumList."""
        ret = SpectrumList(name=self.name)
        for child in self.children:
            ret.children.append(child.__copy__())
        return ret



from PyQt5 import QtCore, QtGui, QtWidgets

from .stylewidget_gui import Ui_Form
from PyQt5.QtWidgets import QColorDialog
from PyQt5.QtGui import QColor
# from PyQt5.QtWidgets import QMessageBox, QLineEdit, QCheckBox
# import numpy as np
#
# from ..spectrum import Spectrum
#
# import pyqtgraph as pg


from PyQt5.QtCore import Qt
from ..settings import Settings


class StyleWidget(QtWidgets.QWidget, Ui_Form):
    # static variables
    is_opened = False
    _instance = None

    def __init__(self, dock_widget, accepted_func, selected_item, parent=None):
        super(StyleWidget, self).__init__(parent)

        self.setupUi(self)

        self.dock_widget = dock_widget
        self.accepted_func = accepted_func
        # self.selected_item = selected_item

        self.color = None
        self.sym_brush_color = None
        self.sym_fill_color = None

        self.btnOK.clicked.connect(self.accept)
        self.btnApply.clicked.connect(self.applied)
        self.btnCancel.clicked.connect(self.reject)
        self.btnSetDefauls.clicked.connect(self.set_defaults)
        self.cbColor.toggled.connect(self.cbColor_checked_changed)
        self.cbLineWidth.toggled.connect(self.cbLineWidth_checked_changed)
        self.cbLineType.toggled.connect(self.cbLineType_checked_changed)
        self.btnColor.clicked.connect(self.pick_color_clicked)
        self.btnSymBrushColor.clicked.connect(self.pick_color_brush_clicked)
        self.btnSymFillColor.clicked.connect(self.pick_color_fill_clicked)

        self.cbSymBrushDefault.toggled.connect(self.cbSymBrushDefault_checked_changed)
        self.cbSymFillDefault.toggled.connect(self.cbSymFillDefault_checked_changed)

        self.line_types = [
            {'name': 'Solid line', 'index': Qt.SolidLine},
            {'name': 'Dashed line', 'index': Qt.DashLine},
            {'name': 'Dotted line', 'index': Qt.DotLine},
            {'name': 'Dash-dotted line', 'index': Qt.DashDotLine},
            {'name': 'Dash-dot-dotted line', 'index': Qt.DashDotDotLine},
            {'name': 'No line', 'index': Qt.NoPen}
        ]

        self.symbol_types = [
            {'name': 'No symbol', 'sym': None},
            {'name': '\u25CF', 'sym': 'o'},
            {'name': '\u25BC', 'sym': 't'},
            {'name': '\u25B2', 'sym': 't1'},
            {'name': '\u25BA', 'sym': 't2'},
            {'name': '\u25C4', 'sym': 't3'},
            {'name': '\u25A0', 'sym': 's'},
            {'name': '\u2B1F', 'sym': 'p'},
            {'name': '\u2B22', 'sym': 'h'},
            {'name': '\u2605', 'sym': 'star'},
            {'name': '+', 'sym': '+'},
            {'name': '\u2666', 'sym': 'd'}
        ]

        self.combLineType.addItems(map(lambda m: m['name'], self.line_types))
        self.combSymbol.addItems(map(lambda m: m['name'], self.symbol_types))

        # self.sbAlpha.setValue(255)
        # self.sbSymBrushAlpha.setValue(255)
        # self.sbSymFillAlpha.setValue(255)

        self.dsbLineWidth.setValue(Settings.line_width)
        self.combLineType.setCurrentIndex(0)
        self.combSymbol.setCurrentIndex(0)

        if not hasattr(selected_item, 'line_alpha'):
            setattr(selected_item, 'line_alpha', 255)

        if selected_item.color is None:
            self.cbColor.setChecked(True)
        else:
            self.color = QColor(*selected_item.color) if isinstance(selected_item.color, (tuple, list)) else QColor(
                selected_item.color)

            self.btnColor.setStyleSheet("QWidget {background-color: %s}" % self.color.name())

        self.sbAlpha.setValue(selected_item.line_alpha)

        if selected_item.line_width is None:
            self.cbLineWidth.setChecked(True)
        else:
            self.dsbLineWidth.setValue(float(selected_item.line_width))

        if selected_item.line_type is None:
            self.cbLineType.setChecked(True)
        else:
            curr_type = list(filter(lambda t: t['index'] == selected_item.line_type, self.line_types))
            self.combLineType.setCurrentIndex(self.line_types.index(curr_type[0]))

        if not hasattr(selected_item, 'symbol'):
            setattr(selected_item, 'symbol', None)

        if not hasattr(selected_item, 'symbol_brush'):
            setattr(selected_item, 'symbol_brush', None)

        if not hasattr(selected_item, 'symbol_fill'):
            setattr(selected_item, 'symbol_fill', None)

        if not hasattr(selected_item, 'symbol_size'):
            setattr(selected_item, 'symbol_size', 8)

        if not hasattr(selected_item, 'sym_brush_alpha'):
            setattr(selected_item, 'sym_brush_alpha', 255)

        if not hasattr(selected_item, 'sym_fill_alpha'):
            setattr(selected_item, 'sym_fill_alpha', 255)

        if selected_item.symbol is None:
            self.combSymbol.setCurrentIndex(0)
        else:
            curr_sym_type = list(filter(lambda t: t['sym'] == selected_item.symbol, self.symbol_types))
            self.combSymbol.setCurrentIndex(self.symbol_types.index(curr_sym_type[0]))

        if selected_item.symbol_brush is None:
            self.cbSymBrushDefault.setChecked(True)
        else:
            self.sym_brush_color = QColor(*selected_item.color) if isinstance(selected_item.symbol_brush, (tuple, list)) else QColor(
                selected_item.symbol_brush)

            self.btnSymBrushColor.setStyleSheet("QWidget {background-color: %s}" % self.sym_brush_color.name())
            self.sbSymBrushAlpha.setValue(self.sym_brush_color.alpha())

        if selected_item.symbol_fill is None:
            self.cbSymFillDefault.setChecked(True)
        else:
            self.sym_fill_color = QColor(*selected_item.color) if isinstance(selected_item.symbol_fill, (tuple, list)) else QColor(
                selected_item.symbol_fill)

            self.btnSymFillColor.setStyleSheet("QWidget {background-color: %s}" % self.sym_fill_color.name())
            self.sbSymFillAlpha.setValue(self.sym_fill_color.alpha())

        self.sbSymBrushAlpha.setValue(selected_item.sym_brush_alpha)
        self.sbSymFillAlpha.setValue(selected_item.sym_fill_alpha)
        self.dsbSymSize.setValue(float(selected_item.symbol_size))

        if not hasattr(selected_item, 'plot_legend'):
            setattr(selected_item, 'plot_legend', True)

        self.cbPlotLegend.setChecked(selected_item.plot_legend)

        self.accepted = False
        StyleWidget.is_opened = True
        StyleWidget._instance = self

        self.dock_widget.parent().resizeDocks([self.dock_widget], [150], Qt.Vertical)
        self.dock_widget.titleBarWidget().setText("Style settings")
        self.dock_widget.setWidget(self)
        self.dock_widget.setVisible(True)

    def cbSymBrushDefault_checked_changed(self):
        checked = self.cbSymBrushDefault.isChecked()
        self.btnSymBrushColor.setEnabled(not checked)

    def cbSymFillDefault_checked_changed(self):
        checked = self.cbSymFillDefault.isChecked()
        self.btnSymFillColor.setEnabled(not checked)

    def cbColor_checked_changed(self):
        checked = self.cbColor.isChecked()
        self.btnColor.setEnabled(not checked)
        # self.sbAlpha.setEnabled(not checked)

    def cbLineWidth_checked_changed(self):
        checked = self.cbLineWidth.isChecked()
        self.dsbLineWidth.setEnabled(not checked)

    def cbLineType_checked_changed(self):
        checked = self.cbLineType.isChecked()
        self.combLineType.setEnabled(not checked)

    def pick_color_clicked(self):
        color = QColorDialog.getColor(self.color if self.color else QColor('white'))
        if color.isValid():
            self.btnColor.setStyleSheet("QWidget { background-color: %s}" % color.name())
            self.color = color

    def pick_color_brush_clicked(self):
        color = QColorDialog.getColor(self.sym_brush_color if self.sym_brush_color else QColor('white'))
        if color.isValid():
            self.btnSymBrushColor.setStyleSheet("QWidget { background-color: %s}" % color.name())
            self.sym_brush_color = color

    def pick_color_fill_clicked(self):
        color = QColorDialog.getColor(self.sym_fill_color if self.sym_fill_color else QColor('white'))
        if color.isValid():
            self.btnSymFillColor.setStyleSheet("QWidget { background-color: %s}" % color.name())
            self.sym_fill_color = color

    def set_defaults(self):
        self.cbColor.setChecked(True)
        self.cbLineType.setChecked(True)
        self.cbLineWidth.setChecked(True)
        self.cbSymBrushDefault.setChecked(True)
        self.cbSymFillDefault.setChecked(True)
        self.combLineType.setCurrentIndex(0)
        self.combSymbol.setCurrentIndex(0)

    def applied(self):
        self.accepted_func()

    def accept(self):
        self.accepted = True
        StyleWidget.is_opened = False
        StyleWidget._instance = None
        self.dock_widget.setVisible(False)
        self.accepted_func()

    def reject(self):
        # self.remove_last_fit()
        StyleWidget.is_opened = False
        StyleWidget._instance = None
        self.dock_widget.setVisible(False)

#
# if __name__ == "__main__":
#     import sys
#
#
#     def my_exception_hook(exctype, value, traceback):
#         # Print the error and traceback
#         print(exctype, value, traceback)
#         # Call the normal Exception hook after
#         sys._excepthook(exctype, value, traceback)
#         sys.exit(1)
#
#
#     from PyQt5.QtWidgets import QApplication
#
#     sys._excepthook = sys.excepthook
#
#     # Set the exception hook to our wrapping function
#     sys.excepthook = my_exception_hook
#
#     app = QtWidgets.QApplication(sys.argv)
#     Dialog = FitWidget(None, None)
#     # Dialog.show()
#     sys.exit(app.exec_())



from PyQt5 import QtCore

from pyqtgraph.graphicsItems.LabelItem import LabelItem as _LabelItem

# changed vertical spacing in legend entries


class LabelItem(_LabelItem):
    """
    GraphicsWidget displaying text.
    Used mainly as axis labels, titles, etc.

    Note: To display text inside a scaled view (ViewBox, PlotWidget, etc) use TextItem
    """

    def __init__(self, text=' ', parent=None, angle=0, verspacing=5, **args):
        self.verspacing = verspacing  # added vertical spacing parameter
        super(LabelItem, self).__init__(text, parent, angle, **args)

    def updateMin(self):
        bounds = self.itemRect()
        self.setMinimumWidth(bounds.width())

        # changed here
        # self.setMinimumHeight(bounds.height())  # commented
        self.setMinimumHeight(self.verspacing)  # added

        self._sizeHint = {
            QtCore.Qt.MinimumSize: (bounds.width(), bounds.height()),
            QtCore.Qt.PreferredSize: (bounds.width(), bounds.height()),
            QtCore.Qt.MaximumSize: (-1, -1),  # bounds.width()*2, bounds.height()*2),
            QtCore.Qt.MinimumDescent: (0, 0)  ##?? what is this?
        }

        self.updateGeometry()



# import warnings
# warnings.filterwarnings("ignore")

import numpy as np
import matplotlib.pyplot as plt
import json
from scipy.integrate import odeint
import re


class GeneralModel:

    def __init__(self):
        self.initial_conditions = {}
        self.elem_reactions = []  # list of dictionaries of elementary reactions
        self.scheme = ""
        self.func = None  # function that returns dc_dt for current c and t for build model

    @classmethod
    def from_text(cls, scheme):
        """
        Expected format is single or multiline, reactions are denoted with equal sign,
        which allows for forward and backward rates. Names of species are case sensitive and can contain numbers.
        Eg. decay of triplet benzophenone with mixed 1st and 2nd order (self TT annihilation),
        here zero represents the ground state BP:
            BP3 = BP_GS  # decay of triplet BP to GS
            2BP3 = BP3 + BP_GS  # selfquenching of triplet BP

        Eg. pandemic SIR model:
            Susceptible + Infected = 2Infected
            Infected = Recovered
            Infected = Dead

        :param scheme:
            input text-based model
        :return:
            Model representing input reaction scheme.
        """

        if scheme.strip() == '':
            raise ValueError("Parameter scheme is empty!")

        _model = cls()
        _model.scheme = scheme
        # find name containing chars and numbers that can start with a number
        pattern = re.compile(r'(\d)([\w\d]+)|([\w\d]+)')

        for line in filter(None, scheme.split('\n')):  # filter removes empty entries from split

            line = list(filter(None, line.split('#')))[0]  # remove comments, take only the characters before possible #
            tokens = []

            for token in filter(None, line.split('=')):  # remove empty entries
                entries = []

                # process possible number in front of species, species cannot have numbers in their text
                for entry in list(filter(None, token.split('+'))):
                    string = ''.join(filter(lambda d: not d.isspace(), list(entry)))  # remove white space chars

                    match = re.match(pattern, string)

                    groups = list(match.groups())

                    # groups[2] is None means we start the name with a number
                    number = int(groups[0]) if groups[2] is None else 1
                    name = groups[1] if groups[2] is None else groups[2]

                    if number < 1:
                        number = 1

                    entries += number * [name]  # list arithmetics

                tokens.append(entries)

            for reactants, products in zip(tokens[:-1], tokens[1:]):
                _model.add_elementary_reaction(reactants, products)

        comps = _model.get_compartments()
        init = [1 if i == 0 else 0 for i in range(len(comps))]
        _model.initial_conditions = dict(zip(comps, init))

        return _model

    def add_elementary_reaction(self, from_comp=('A', 'A'), to_comp=('B', 'C'), forward_rate=1, backward_rate=0):
        from_comp = from_comp if isinstance(from_comp, (list, tuple)) else [from_comp]
        to_comp = to_comp if isinstance(to_comp, (list, tuple)) else [to_comp]

        reaction = dict(from_comp=from_comp, to_comp=to_comp, forward_rate=forward_rate, backward_rate=backward_rate)

        for el in self.elem_reactions:  # replace rate if the same elementary reaction is already in list
            if el['from_comp'] == reaction['from_comp'] and el['to_comp'] == reaction['to_comp']:
                el['forward_rate'] = reaction['forward_rate']
                el['backward_rate'] = reaction['backward_rate']
                return

        self.elem_reactions.append(reaction)

    def get_compartments(self):
        """
        Return the compartment names, the names are case sensitive.
        """
        names = []
        for el in self.elem_reactions:
            for c in el['from_comp']:
                if c not in names:
                    names.append(c)

            for c in el['to_comp']:
                if c not in names:
                    names.append(c)
        return names

    def build_func(self):
        """
        Builds model and returns the function that takes c, t and rates as an argument
        and can be directly used for odeint numerical integration method.
        """

        comps = self.get_compartments()

        idx_dict = dict(enumerate(comps))
        inv_idx = dict(zip(idx_dict.values(), idx_dict.keys()))

        r = len(self.elem_reactions)
        idx_from = []  # arrays of arrays of indexes for each elementary reaction
        idx_to = []  # arrays of arrays of indexes for each elementary reaction
        _rates = np.empty((r, 2), dtype=np.float64)

        # build the lists of indexes and so on...
        for i, el in enumerate(self.elem_reactions):
            i_from = list(map(lambda com: inv_idx[com], el['from_comp']))  # list of indexes of starting materials
            i_to = list(map(lambda com: inv_idx[com], el['to_comp']))  # list of indexes of reaction products

            idx_from.append(i_from)
            idx_to.append(i_to)

            _rates[i, 0] = el['forward_rate']
            _rates[i, 1] = el['backward_rate']

        # TODO: possible space for optimization if found too slow for odeint, probably some Cython or C code would be
        # TODO: needed
        def func(c, t, rates=None):
            """Rates if provided must be (r x 2) matrix, where in first column are forward rates and second column
            backward rates. r is number of elementary reactions."""

            rates = _rates if rates is None else rates

            dc_dt = np.zeros_like(c)

            for i in range(r):
                forward_prod, backward_prod = rates[i]

                # eg. for elementary step A + B = C + D
                for k in idx_from[i]:
                    forward_prod *= c[k]  # forward rate products, eg. k_AB * [A] * [B]

                for k in idx_to[i]:
                    backward_prod *= c[k]  # backward rate products, eg. k_CD * [C] * [D]

                for k in idx_from[i]:
                    dc_dt[k] += backward_prod - forward_prod  # reactants

                for k in idx_to[i]:
                    dc_dt[k] += forward_prod - backward_prod  # products

            return dc_dt

        self.func = func
        return func

    def simulate_model(self, times):
        comp = self.get_compartments()
        j = np.asarray(list(self.initial_conditions.values()))  # assuming the order is the same as form self.get_compartments

        solution = odeint(self.build_func(), j, times)

        plt.xlabel('Time')
        plt.ylabel('Concentration')

        for i, c in enumerate(comp):
            plt.plot(times, solution[:, i], label=f'{c}')

        plt.legend(frameon=False)
        plt.show()

    def get_rates(self, get_backwards_rates=True, append_values=True):
        rates = []

        # for el in self.elem_reactions:
        #     name = f"k({','.join(el['from_comp'])}\u2192{','.join(el['to_comp'])})"
        #     rates.append([name, el['forward_rate']] if append_values else name)
        #
        #     if get_backwards_rates:
        #         name = f"k({','.join(el['to_comp'])}\u2192{','.join(el['from_comp'])})"
        #         rates.append([name, el['backward_rate']] if append_values else name)

        for el in self.elem_reactions:
            name = f"k_{'_'.join(el['from_comp'])}_{'_'.join(el['to_comp'])}"
            rates.append([name, el['forward_rate']] if append_values else name)

            if get_backwards_rates:
                name = f"k_{'_'.join(el['to_comp'])}_{'_'.join(el['from_comp'])}"
                rates.append([name, el['backward_rate']] if append_values else name)

        return rates

    def set_rates(self, rate_matrix):
        """rate_matrix is a (r x 2) matrix where r is number of elementary reactions.
        first column is forward rate, second backward rate. rate_matrix can be list of
        lists."""

        assert len(rate_matrix) == len(self.elem_reactions)

        for el, (f_rate, b_rate) in zip(self.elem_reactions, rate_matrix):
            el['forward_rate'] = f_rate
            el['backward_rate'] = b_rate

    @classmethod
    def load(cls, fpath='general models/mod1.json'):
        _model = cls()

        try:
            with open(fpath, "r") as file:
                _model.scheme, _model.initial_conditions, _model.elem_reactions = json.load(file).values()

        except Exception as ex:
            print('Error loading model:\n' + ex.__str__())

        return _model

    def save(self, fpath='general models/mod1.json'):
        try:
            with open(fpath, "w") as file:

                to_dump = dict(scheme=self.scheme,
                               initial_conditions=self.initial_conditions,
                               elementary_reactions=self.elem_reactions)

                json.dump(to_dump, file, sort_keys=False, indent=4, separators=(',', ': '))

        except Exception as ex:
            print('Error saving model:\n' + ex.__str__())

    def print_model(self):
        print(f'Scheme: {self.scheme}')
        print(f'Initial conditions: {self.initial_conditions}')

        for el in self.elem_reactions:
            print(f"Elementary reaction: {' + '.join(el['from_comp'])} \u2192 {' + '.join(el['to_comp'])}, "
                  f"forward_rate: {el['forward_rate']}, backward_rate: {el['backward_rate']}")


# def main():
#     SIR = 'S + I = I + E\nE = I = R = S'
#
#     # SIR = 'BP3 = BP_GS, 2BP3 = BP3 + BP_GS'
#
#     model = GeneralModel.from_text(SIR)
#     model.initial_conditions['I'] = 0.0001
#     model.initial_conditions['S'] = 1
#
#     #
#     model.elem_reactions[0]['forward_rate'] = 2
#     model.elem_reactions[1]['forward_rate'] = 0.1
#     model.elem_reactions[2]['forward_rate'] = 0.2
#     model.elem_reactions[3]['forward_rate'] = 0.01
#
#     #
#     # model.elem_reactions[0]['backward_rate'] = 0.2
#     # model.elem_reactions[1]['backward_rate'] = 0.7
#     # model.elem_reactions[2]['backward_rate'] = 0.5
#     # model.elem_reactions[3]['forward_rate'] = 0.1
#
#     model.print_model()
#
#     #
#     # model.add_elementary_reaction(['Susceptible', 'Infected'], ['Infected', 'Infected'], 0.5)
#     # model.add_elementary_reaction('Infected', 'Recovered', 0.1)
#     # model.add_elementary_reaction('Infected', 'Dead', 0.01)
#
#     # model.add_elementary_reaction(['A', 'B'], ['B'], 0.5)
#     # model.add_elementary_reaction('Infected', 'Recovered', 0.1)
#     # model.add_elementary_reaction('Infected', 'Dead', 0.01)
#     # model.add_elementary_reaction('Recovered', 'Susceptible', 0.01)
#
#     times = np.linspace(0, 200, 1000, dtype=np.float64)
#
#     model.simulate_model(times)
#
#     model.save(fpath='general models/Pandemic SEIRS.json')

#
# if __name__ == '__main__':
#
#     main()

    # reaction = 'A + B = C = D = E\n A + E=B'
    # reaction = 'BP = zero, 2 BP = zero + BP'
    # SIR = 'Susceptible + Infected = 2 Infected, Infected = Recovered, Infected = Dead'
    #
    # model = GeneralModel.from_text(SIR)
    # print(model.get_compartments())
    # model.print_model()
    #
    # model.elem_reactions[1]['forward_rate'] = 0.1
    # model.elem_reactions[2]['forward_rate'] = 0.01
    #
    # times = np.linspace(0, 50, 1000, dtype=np.float64)
    #
    # model.simulate_model(times, [1, 0.01, 0, 0])

    # cProfile.run('main()')
    # model = TargetModel.load()
    # model.print_model()
    # print(model.get_rate_names())
    # model.plot_model()
    #
    # model = GeneralModel()
    #
    # # compartments = list('abcdefghijk')
    # # n = len(compartments)
    # # for i in range(n):
    # #     j = (i + 1) % n
    # #     k = (i + 2) % n
    # #     model.add_transition(compartments[i], compartments[j], i*j + 1)
    # #     # model.add_transition(compartments[i], compartments[k], i*k + 1)
    # #
    # model.add_elementary_reaction(['Susceptible', 'Infected'], ['Infected', 'Infected'], 0.5)
    # model.add_elementary_reaction('Infected', 'Recovered', 0.1)
    # model.add_elementary_reaction('Infected', 'Dead', 0.01)
    #
    # # model.add_elementary_reaction('A', ['B', 'C'], 1, 1)
    # # # model.add_elementary_reaction(['A', 'B'], 'C', 1, 0.01)
    # # model.add_elementary_reaction(['C', 'A'], 'B', 1, 0)
    #
    # times = np.linspace(0, 80, 1000, dtype=np.float64)
    #
    # # model.print_model()
    # model.simulate_model(times, [1, 0.001, 0, 0])

    # print(model.get_compartments())
    # print(model.get_rate_names())





# from spectramanipulator.spectrum import Spectrum, SpectrumList
# from spectramanipulator.settings import Settings
# from spectramanipulator.logger import Logger
# from spectramanipulator.console import Console
#
# __all__ = ['Spectrum', 'SpectrumList', 'Settings', 'Logger', 'Console']
#


#  Spectra Manipulator version
__version__ = "0.1.1"





from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QFileDialog, QListView, QAbstractItemView, QTreeView, QListWidget, QMessageBox

from PyQt5.QtCore import Qt
from PyQt5 import QtGui
from spectramanipulator.dialogs.gui_load_kinetics import Ui_Dialog
from spectramanipulator.settings import Settings
import os


class MyQListWidget(QListWidget):

    def __init__(self, parent=None):
        super(MyQListWidget, self).__init__(parent=parent)
        self.setSelectionMode(QAbstractItemView.MultiSelection)

        self.item_names = []  # shortcut paths
        self.paths = []  # directory paths

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        if e.key() == 16777223:  # del pressed
            for item in self.selectedItems():
                index = self.item_names.index(item.text())
                del self.item_names[index]
                del self.paths[index]
                self.takeItem(self.row(item))

    def addItem(self, aitem, path) -> None:
        super(MyQListWidget, self).addItem(aitem)
        self.item_names.append(aitem)
        self.paths.append(path)


class LoadKineticsDialog(QtWidgets.QDialog, Ui_Dialog):

    # static variables
    is_opened = False
    _instance = None

    def __init__(self, parent=None):
        super(LoadKineticsDialog, self).__init__(parent)
        self.setupUi(self)

        self.lwFolders = MyQListWidget(self)

        #disable resizing of the window,
        # help from https://stackoverflow.com/questions/16673074/in-qt-c-how-can-i-fully-disable-resizing-a-window-including-the-resize-icon-w
        self.setWindowFlags(Qt.Dialog | Qt.MSWindowsFixedSizeDialogHint)

        self.setWindowTitle("Batch Load Kinetics")

        self.cbKineticsMeasuredByEach.toggled.connect(self.cbKineticsMeasuredByEach_toggled)
        self.cbBCorr.toggled.connect(self.cbBCorr_toggled)
        self.cbCut.toggled.connect(self.cbCut_toggled)
        self.btnChooseDirs.clicked.connect(self.btnChooseDirs_clicked)

        self.cbKineticsMeasuredByEach_toggled()
        self.cbBCorr_toggled()
        self.cbCut_toggled()

        self.lwGridLayout.addWidget(self.lwFolders, 0, 0, 0, 0)

        self.accepted = False

        LoadKineticsDialog.is_opened = True
        LoadKineticsDialog._instance = self

        self.show()
        self.exec()

    @staticmethod
    def get_instance():
        return LoadKineticsDialog._instance

    @staticmethod
    def check_state(checked):
        return Qt.Checked if checked else 0

    def btnChooseDirs_clicked(self):
        # https://stackoverflow.com/questions/38252419/how-to-get-qfiledialog-to-select-and-return-multiple-folders
        # just copied :)
        file_dialog = QFileDialog()
        file_dialog.setFileMode(QFileDialog.DirectoryOnly)
        file_dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        file_dialog.setDirectory(Settings.load_kinetics_last_path)
        file_view = file_dialog.findChild(QListView, 'listView')

        # to make it possible to select multiple directories:
        if file_view:
            file_view.setSelectionMode(QAbstractItemView.MultiSelection)
        f_tree_view = file_dialog.findChild(QTreeView)
        if f_tree_view:
            f_tree_view.setSelectionMode(QAbstractItemView.MultiSelection)

        if file_dialog.exec():
            for path in file_dialog.selectedFiles():
                head, tail = os.path.split(path)
                head2, tail2 = os.path.split(head)
                name = os.path.join(tail2, tail)
                Settings.load_kinetics_last_path = head
                if name not in self.lwFolders.item_names:
                    self.lwFolders.addItem(name, path)

    def cbBCorr_toggled(self):
        checked = self.cbBCorr.isChecked()
        self.leBCorr0.setEnabled(checked)
        self.leBCorr1.setEnabled(checked)

    def cbCut_toggled(self):
        checked = self.cbCut.isChecked()
        self.leCut0.setEnabled(checked)
        self.leCut1.setEnabled(checked)

    def cbKineticsMeasuredByEach_toggled(self):
        checked = self.cbKineticsMeasuredByEach.isChecked()
        self.leTimeUnit.setEnabled(checked)
        self.leTimes.setEnabled(not checked)

    def accept(self):
        try:
            float(self.leTimeUnit.text())
            float(self.leCut0.text())
            float(self.leCut1.text())
            float(self.leBCorr0.text())
            float(self.leBCorr1.text())
        except ValueError:
            QMessageBox.critical(self, 'Error', "Invalid input, please check the fields.")
            return

        self.accepted = True
        LoadKineticsDialog.is_opened = False
        LoadKineticsDialog._instance = None
        Settings.save()
        super(LoadKineticsDialog, self).accept()

    def reject(self):
        LoadKineticsDialog.is_opened = False
        LoadKineticsDialog._instance = None
        super(LoadKineticsDialog, self).reject()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = LoadKineticsDialog()
    # Dialog.show()
    sys.exit(app.exec_())







from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtCore import Qt


class CheckBoxRC(QCheckBox):
    """Classical checkbox that allows right click which toggles the checkbox the same as left click,
    if checkbox was right clicked, the attribute self.right_button_pressed will be set to True."""

    def __init__(self, text: str = '', parent=None):
        super(CheckBoxRC, self).__init__(text, parent)

        self.right_button_pressed = False

    def mousePressEvent(self, e: QMouseEvent):
        if e.button() == Qt.LeftButton:
            self.right_button_pressed = False
            super(CheckBoxRC, self).mousePressEvent(e)
        elif e.button() == Qt.RightButton:
            self.right_button_pressed = True
            self.setChecked(not self.isChecked())





import numpy as np
from spectramanipulator.spectrum import Spectrum
from spectramanipulator.parsers.parser import Parser


class DXFileParser(Parser):

    def __init__(self, filepath=None, str_data=None, delimiter=' ', decimal_sep='.',
                 dx_import_spectra_name_from_filename=False, dx_if_title_is_empty_use_filename=True):
        super(DXFileParser, self).__init__(filepath, str_data, delimiter, decimal_sep)

        self.dx_import_spectra_name_from_filename = dx_import_spectra_name_from_filename
        self.dx_if_title_is_empty_use_filename = dx_if_title_is_empty_use_filename

        self.buffer = []

    def _parse_data(self, data, name):

        if len(data) < 2:
            data.clear()
            return

        sp = Spectrum(np.asarray(data, dtype=np.float64), filepath=self.filepath, name=name)
        self.buffer.append(sp)
        data.clear()

    def parse(self, name=''):

        it = self.__iter__()  # get line iterator

        data = []

        # read first header title line
        title_header = it.__next__().strip()
        if name == '':
            try:
                split = title_header.split('=')
                name = split[1][1:]
            except:
                pass

        # if the title field in the dx file is empty, or some error occurred while reading it,
        # use user defined settings

        if self.dx_import_spectra_name_from_filename and not self.dx_if_title_is_empty_use_filename:
            name = self.name_of_file

        if self.dx_if_title_is_empty_use_filename:
            name = self.name_of_file if name == '' else name

        # skip 17 lines of header
        # for i in range(17):
        #     it.__next__()

        # test_split = it.__next__().strip()
        # if self.float_try_parse(test_split[0]) is None:
        #     for i in range(6):
        #         it.__next__()

        # count = 0

        # read the data itself
        for line in it:
            split = line.split(self.delimiter if self.delimiter != '' else None)
            if len(split) < 2:
                self._parse_data(data, name)
                continue

            wavelength = self.float_try_parse(split[0])
            absorbance = self.float_try_parse(split[1])
            if wavelength is None or absorbance is None:
                self._parse_data(data, name)
                continue

            data.append((wavelength, absorbance))

        self._parse_data(data, name)

        # one row will be skipped every time - it is the end line of DX file
        # if count > 1:
        #     Logger.console_message("Warning, {} rows were skipped due to unsuccessful "
        #                            "float value parsing of DX file. Some data may be lost.\nFile: {}".format(count - 1, self.filepath))

        # sp = Spectrum(np.asarray(data, dtype=np.float64), self.filepath, name)
        # return [sp]  # returns a list

        return self.buffer



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'spectramanipulator/dialogs\gui_fit_widget_old.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(444, 591)
        self.gridLayout_4 = QtWidgets.QGridLayout(Form)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.tabWidget = QtWidgets.QTabWidget(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.gridLayout = QtWidgets.QGridLayout(self.tab)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_2 = QtWidgets.QLabel(self.tab)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_4.addWidget(self.label_2)
        self.cbModel = QtWidgets.QComboBox(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbModel.sizePolicy().hasHeightForWidth())
        self.cbModel.setSizePolicy(sizePolicy)
        self.cbModel.setObjectName("cbModel")
        self.horizontalLayout_4.addWidget(self.cbModel)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.cbCustom = QtWidgets.QCheckBox(self.tab)
        self.cbCustom.setObjectName("cbCustom")
        self.verticalLayout.addWidget(self.cbCustom)
        self.label = QtWidgets.QLabel(self.tab)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.pteEquation = QtWidgets.QPlainTextEdit(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(4)
        sizePolicy.setHeightForWidth(self.pteEquation.sizePolicy().hasHeightForWidth())
        self.pteEquation.setSizePolicy(sizePolicy)
        self.pteEquation.setMinimumSize(QtCore.QSize(0, 0))
        self.pteEquation.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.pteEquation.setPlainText("")
        self.pteEquation.setObjectName("pteEquation")
        self.verticalLayout.addWidget(self.pteEquation)
        self.cbUpdateInitValues = QtWidgets.QCheckBox(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbUpdateInitValues.sizePolicy().hasHeightForWidth())
        self.cbUpdateInitValues.setSizePolicy(sizePolicy)
        self.cbUpdateInitValues.setObjectName("cbUpdateInitValues")
        self.verticalLayout.addWidget(self.cbUpdateInitValues)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_9 = QtWidgets.QLabel(self.tab)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_3.addWidget(self.label_9)
        self.sbParamsCount = QtWidgets.QSpinBox(self.tab)
        self.sbParamsCount.setMinimum(1)
        self.sbParamsCount.setMaximum(10)
        self.sbParamsCount.setObjectName("sbParamsCount")
        self.horizontalLayout_3.addWidget(self.sbParamsCount)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.scrollArea_2 = QtWidgets.QScrollArea(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(10)
        sizePolicy.setHeightForWidth(self.scrollArea_2.sizePolicy().hasHeightForWidth())
        self.scrollArea_2.setSizePolicy(sizePolicy)
        self.scrollArea_2.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea_2.setWidgetResizable(True)
        self.scrollArea_2.setObjectName("scrollArea_2")
        self.scrollAreaWidgetContents_3 = QtWidgets.QWidget()
        self.scrollAreaWidgetContents_3.setGeometry(QtCore.QRect(0, 0, 388, 86))
        self.scrollAreaWidgetContents_3.setObjectName("scrollAreaWidgetContents_3")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents_3)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.predefGridLayout = QtWidgets.QGridLayout()
        self.predefGridLayout.setObjectName("predefGridLayout")
        self.label_7 = QtWidgets.QLabel(self.scrollAreaWidgetContents_3)
        self.label_7.setAlignment(QtCore.Qt.AlignCenter)
        self.label_7.setObjectName("label_7")
        self.predefGridLayout.addWidget(self.label_7, 0, 4, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.scrollAreaWidgetContents_3)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.predefGridLayout.addWidget(self.label_3, 0, 0, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.scrollAreaWidgetContents_3)
        self.label_8.setAlignment(QtCore.Qt.AlignCenter)
        self.label_8.setObjectName("label_8")
        self.predefGridLayout.addWidget(self.label_8, 0, 5, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.scrollAreaWidgetContents_3)
        self.label_6.setAlignment(QtCore.Qt.AlignCenter)
        self.label_6.setObjectName("label_6")
        self.predefGridLayout.addWidget(self.label_6, 0, 3, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.scrollAreaWidgetContents_3)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.predefGridLayout.addWidget(self.label_4, 0, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.scrollAreaWidgetContents_3)
        self.label_5.setAlignment(QtCore.Qt.AlignCenter)
        self.label_5.setObjectName("label_5")
        self.predefGridLayout.addWidget(self.label_5, 0, 2, 1, 1)
        self.gridLayout_6.addLayout(self.predefGridLayout, 0, 0, 1, 1)
        self.scrollArea_2.setWidget(self.scrollAreaWidgetContents_3)
        self.verticalLayout.addWidget(self.scrollArea_2)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.tab_2)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_21 = QtWidgets.QLabel(self.tab_2)
        self.label_21.setObjectName("label_21")
        self.horizontalLayout_5.addWidget(self.label_21)
        self.cbGenModels = QtWidgets.QComboBox(self.tab_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbGenModels.sizePolicy().hasHeightForWidth())
        self.cbGenModels.setSizePolicy(sizePolicy)
        self.cbGenModels.setObjectName("cbGenModels")
        self.horizontalLayout_5.addWidget(self.cbGenModels)
        self.btnSaveCustomModel = QtWidgets.QPushButton(self.tab_2)
        self.btnSaveCustomModel.setObjectName("btnSaveCustomModel")
        self.horizontalLayout_5.addWidget(self.btnSaveCustomModel)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)
        self.label_22 = QtWidgets.QLabel(self.tab_2)
        self.label_22.setObjectName("label_22")
        self.verticalLayout_2.addWidget(self.label_22)
        self.pteScheme = QtWidgets.QPlainTextEdit(self.tab_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Ignored)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(4)
        sizePolicy.setHeightForWidth(self.pteScheme.sizePolicy().hasHeightForWidth())
        self.pteScheme.setSizePolicy(sizePolicy)
        self.pteScheme.setMinimumSize(QtCore.QSize(0, 0))
        self.pteScheme.setMaximumSize(QtCore.QSize(16777215, 300))
        self.pteScheme.setPlainText("")
        self.pteScheme.setObjectName("pteScheme")
        self.verticalLayout_2.addWidget(self.pteScheme)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.btnBuildModel = QtWidgets.QPushButton(self.tab_2)
        self.btnBuildModel.setObjectName("btnBuildModel")
        self.horizontalLayout_7.addWidget(self.btnBuildModel)
        self.cbShowBackwardRates = QtWidgets.QCheckBox(self.tab_2)
        self.cbShowBackwardRates.setChecked(False)
        self.cbShowBackwardRates.setObjectName("cbShowBackwardRates")
        self.horizontalLayout_7.addWidget(self.cbShowBackwardRates)
        self.verticalLayout_2.addLayout(self.horizontalLayout_7)
        self.cbPlotAllComps = QtWidgets.QCheckBox(self.tab_2)
        self.cbPlotAllComps.setObjectName("cbPlotAllComps")
        self.verticalLayout_2.addWidget(self.cbPlotAllComps)
        self.scrollArea = QtWidgets.QScrollArea(self.tab_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(10)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustIgnored)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents_2 = QtWidgets.QWidget()
        self.scrollAreaWidgetContents_2.setGeometry(QtCore.QRect(0, 0, 388, 134))
        self.scrollAreaWidgetContents_2.setObjectName("scrollAreaWidgetContents_2")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents_2)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.cusModelGridLayout = QtWidgets.QGridLayout()
        self.cusModelGridLayout.setObjectName("cusModelGridLayout")
        self.label_17 = QtWidgets.QLabel(self.scrollAreaWidgetContents_2)
        self.label_17.setAlignment(QtCore.Qt.AlignCenter)
        self.label_17.setObjectName("label_17")
        self.cusModelGridLayout.addWidget(self.label_17, 0, 5, 1, 1)
        self.label_15 = QtWidgets.QLabel(self.scrollAreaWidgetContents_2)
        self.label_15.setAlignment(QtCore.Qt.AlignCenter)
        self.label_15.setObjectName("label_15")
        self.cusModelGridLayout.addWidget(self.label_15, 0, 4, 1, 1)
        self.label_20 = QtWidgets.QLabel(self.scrollAreaWidgetContents_2)
        self.label_20.setAlignment(QtCore.Qt.AlignCenter)
        self.label_20.setObjectName("label_20")
        self.cusModelGridLayout.addWidget(self.label_20, 0, 2, 1, 1)
        self.label_18 = QtWidgets.QLabel(self.scrollAreaWidgetContents_2)
        self.label_18.setAlignment(QtCore.Qt.AlignCenter)
        self.label_18.setObjectName("label_18")
        self.cusModelGridLayout.addWidget(self.label_18, 0, 3, 1, 1)
        self.label_16 = QtWidgets.QLabel(self.scrollAreaWidgetContents_2)
        self.label_16.setAlignment(QtCore.Qt.AlignCenter)
        self.label_16.setObjectName("label_16")
        self.cusModelGridLayout.addWidget(self.label_16, 0, 0, 1, 1)
        self.label_19 = QtWidgets.QLabel(self.scrollAreaWidgetContents_2)
        self.label_19.setAlignment(QtCore.Qt.AlignCenter)
        self.label_19.setObjectName("label_19")
        self.cusModelGridLayout.addWidget(self.label_19, 0, 1, 1, 1)
        self.gridLayout_5.addLayout(self.cusModelGridLayout, 1, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents_2)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.gridLayout_3.addLayout(self.verticalLayout_2, 0, 0, 1, 1)
        self.tabWidget.addTab(self.tab_2, "")
        self.verticalLayout_3.addWidget(self.tabWidget)
        self.label_10 = QtWidgets.QLabel(Form)
        self.label_10.setObjectName("label_10")
        self.verticalLayout_3.addWidget(self.label_10)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_11 = QtWidgets.QLabel(Form)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout.addWidget(self.label_11)
        self.leX0 = MyLineEdit(Form)
        self.leX0.setObjectName("leX0")
        self.horizontalLayout.addWidget(self.leX0)
        self.label_12 = QtWidgets.QLabel(Form)
        self.label_12.setObjectName("label_12")
        self.horizontalLayout.addWidget(self.label_12)
        self.leX1 = MyLineEdit(Form)
        self.leX1.setObjectName("leX1")
        self.horizontalLayout.addWidget(self.leX1)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_13 = QtWidgets.QLabel(Form)
        self.label_13.setObjectName("label_13")
        self.horizontalLayout_6.addWidget(self.label_13)
        self.cbMethod = QtWidgets.QComboBox(Form)
        self.cbMethod.setObjectName("cbMethod")
        self.horizontalLayout_6.addWidget(self.cbMethod)
        self.verticalLayout_3.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.cbPropWeighting = QtWidgets.QCheckBox(Form)
        self.cbPropWeighting.setChecked(False)
        self.cbPropWeighting.setObjectName("cbPropWeighting")
        self.horizontalLayout_2.addWidget(self.cbPropWeighting)
        self.label_14 = QtWidgets.QLabel(Form)
        self.label_14.setObjectName("label_14")
        self.horizontalLayout_2.addWidget(self.label_14)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Ignored)
        self.verticalLayout_3.addItem(spacerItem1)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.btnPrintReport = QtWidgets.QPushButton(Form)
        self.btnPrintReport.setCheckable(False)
        self.btnPrintReport.setChecked(False)
        self.btnPrintReport.setAutoRepeat(False)
        self.btnPrintReport.setAutoDefault(False)
        self.btnPrintReport.setObjectName("btnPrintReport")
        self.gridLayout_2.addWidget(self.btnPrintReport, 1, 1, 1, 1)
        self.btnCancel = QtWidgets.QPushButton(Form)
        self.btnCancel.setAutoDefault(False)
        self.btnCancel.setObjectName("btnCancel")
        self.gridLayout_2.addWidget(self.btnCancel, 1, 4, 1, 1)
        self.btnPlotFunction = QtWidgets.QPushButton(Form)
        self.btnPlotFunction.setCheckable(False)
        self.btnPlotFunction.setChecked(False)
        self.btnPlotFunction.setAutoRepeat(False)
        self.btnPlotFunction.setAutoDefault(False)
        self.btnPlotFunction.setObjectName("btnPlotFunction")
        self.gridLayout_2.addWidget(self.btnPlotFunction, 0, 0, 1, 1)
        self.btnFit = QtWidgets.QPushButton(Form)
        self.btnFit.setCheckable(False)
        self.btnFit.setChecked(False)
        self.btnFit.setAutoRepeat(False)
        self.btnFit.setAutoDefault(False)
        self.btnFit.setObjectName("btnFit")
        self.gridLayout_2.addWidget(self.btnFit, 1, 0, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem2, 1, 2, 1, 1)
        self.btnOK = QtWidgets.QPushButton(Form)
        self.btnOK.setAutoDefault(False)
        self.btnOK.setObjectName("btnOK")
        self.gridLayout_2.addWidget(self.btnOK, 0, 4, 1, 1)
        self.btnClearPlot = QtWidgets.QPushButton(Form)
        self.btnClearPlot.setCheckable(False)
        self.btnClearPlot.setChecked(False)
        self.btnClearPlot.setAutoRepeat(False)
        self.btnClearPlot.setAutoDefault(False)
        self.btnClearPlot.setObjectName("btnClearPlot")
        self.gridLayout_2.addWidget(self.btnClearPlot, 0, 1, 1, 1)
        self.verticalLayout_3.addLayout(self.gridLayout_2)
        self.gridLayout_4.addLayout(self.verticalLayout_3, 0, 0, 1, 1)

        self.retranslateUi(Form)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_2.setText(_translate("Form", "Model:"))
        self.cbCustom.setText(_translate("Form", "Custom model (equation window will be editable)"))
        self.label.setText(_translate("Form", "<html><head/><body><p><span style=\" font-weight:600;\">def func(x, *params):</span></p></body></html>"))
        self.cbUpdateInitValues.setText(_translate("Form", "Update initial values when linear region move is finished"))
        self.label_9.setText(_translate("Form", "Params:"))
        self.label_7.setText(_translate("Form", "Fixed"))
        self.label_3.setText(_translate("Form", "Param"))
        self.label_8.setText(_translate("Form", "Error"))
        self.label_6.setText(_translate("Form", "Upper\n"
"bound"))
        self.label_4.setText(_translate("Form", "Lower\n"
"bound"))
        self.label_5.setText(_translate("Form", "<= Value <="))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("Form", "Predefined Model"))
        self.label_21.setText(_translate("Form", "Saved models:"))
        self.btnSaveCustomModel.setText(_translate("Form", "Save Model As"))
        self.label_22.setText(_translate("Form", "<html><head/><body><p><span style=\" font-weight:600;\">Reaction model:</span></p></body></html>"))
        self.btnBuildModel.setText(_translate("Form", "Build Model"))
        self.cbShowBackwardRates.setText(_translate("Form", "Non-zero backward rates"))
        self.cbPlotAllComps.setText(_translate("Form", "Plot all compartments"))
        self.label_17.setText(_translate("Form", "Error"))
        self.label_15.setText(_translate("Form", "Fixed"))
        self.label_20.setText(_translate("Form", "<= Value <="))
        self.label_18.setText(_translate("Form", "Upper\n"
"bound"))
        self.label_16.setText(_translate("Form", "Param"))
        self.label_19.setText(_translate("Form", "Lower\n"
"bound"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("Form", "Custom Kinetic Model"))
        self.label_10.setText(_translate("Form", "Selected x range:"))
        self.label_11.setText(_translate("Form", "from"))
        self.label_12.setText(_translate("Form", "to"))
        self.label_13.setText(_translate("Form", "<html><head/><body><p>Minimizing algorithm:</p></body></html>"))
        self.cbPropWeighting.setText(_translate("Form", "Proportional weighting"))
        self.label_14.setText(_translate("Form", "e<sub>i</sub>*=1/y<sub>i</sub><sup>2</sup>"))
        self.btnPrintReport.setText(_translate("Form", "Print report"))
        self.btnCancel.setText(_translate("Form", "Cancel"))
        self.btnPlotFunction.setText(_translate("Form", "Plot Function"))
        self.btnFit.setText(_translate("Form", "Fit"))
        self.btnOK.setText(_translate("Form", "Ok"))
        self.btnClearPlot.setText(_translate("Form", "Clear Plot"))
from dialogs.mylineedit import MyLineEdit


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())



from spectramanipulator.fitting.fitmodels import _Model
from scipy.integrate import odeint
import numpy as np


#
# # uncomment this section and change name of the class of new model
# class NewModel(_Model):
#
#     # change name of the model here
#     name = 'User defined model'
#
#     # add corresponding lines to this method for each variable that this model contains
#     # setup the initial value, minimum, maximum and initial option if the variable
#     # will be varied during fitting
#     # use np.inf for positive infinity and -np.inf for negative infinity
#     def init_params(self):
#         self.params.add('c0', value=1, min=0, max=np.inf, vary=False)
#         self.params.add('A', value=1, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('B', value=1, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('C', value=0, min=-np.inf, max=np.inf, vary=False)
#         self.params.add('D', value=0, min=-np.inf, max=np.inf, vary=False)
#         self.params.add('k1', value=1, min=0, max=np.inf, vary=False)
#         self.params.add('k2', value=0.5, min=0, max=np.inf, vary=False)
#         self.params.add('k3', value=1, min=0, max=np.inf, vary=True)
#         self.params.add('k4', value=1, min=0, max=np.inf, vary=True)
#         self.params.add('y0', value=0, min=-np.inf, max=np.inf, vary=True)
#
#     # change the procedure to how to initialize the values of your parameters from the selected user data
#     # the x_data is array of x values region of fitted data and y_data is array of y values of fitted data
#     # this is not mandatory, if you want, you can omit the initilization, it this case, just delete all lines
#     # and keep only 'pass'
#     def initialize_values(self, x_data, y_data):
#         self.params['A'].value = np.max(y_data[0])
#         # self.params['B'].value = y_data[0]
#
#         pass
#
#     # change this function, the order of the parameters in the func method must be the same as was defined in
#     # init_params method!, first is parameter x, which are the x values of fitted data, then the parameters.
#     # the function has to always return numpy array that is the same dimension as x
#     # for solving differential equation, use odeint method
#     @staticmethod
#     def func(x, c0, A, B, C, D, k1, k2, k3, y0):
#         def concB(k1, k2, x):
#             # if k1 == k2, the integrated law fow cB has to be changed
#             if np.abs(k1 - k2) < 1e-10:
#                 return k1 * x * np.exp(-k1 * x)
#             else:  # we would get zero division error for k1 = k2
#                 return (k1 / (k2 - k1)) * (np.exp(-k1 * x) - np.exp(-k2 * x))
#
#         def dC_dt(cC, t):
#             cB = concB(k1, k2, t)
#             return k2 * cB - k3 * cC  # d[C]/dt = k2[B] - k3[C]
#
#         # numerically integrate, initial condition, cC(t=0) = 0
#         # but initial point for odeint must be defined for x[0]
#         # evolve from 0 to x[0] and then take the last point as the initial condition
#         x0 = np.linspace(0, x[0], num=100)
#         init_C = odeint(dC_dt, 0, x0).flatten()[-1]
#         result = odeint(dC_dt, init_C, x)
#
#         cA = np.exp(-k1 * x)
#         cB = concB(k1, k2, x)
#         cC = result.flatten()
#         cD = 1 - cA - cB - cC
#
#         return A * cA + B * cB + C * cC + D * cD + y0

class DF(_Model):
    """l"""

    name = 'DF plot'

    def init_params(self):
        self.params.add('Xan', value=0, min=-np.inf, max=np.inf, vary=False)
        self.params.add('DR', value=0, min=-np.inf, max=np.inf, vary=False)
        self.params.add('DF', value=1, min=-np.inf, max=np.inf, vary=True)

        self.params.add('A0', value=0, min=0, max=10, vary=False)
        self.params.add('kex', value=1.5, min=0, max=np.inf, vary=True)
        self.params.add('k_ex', value=0, min=0, max=np.inf, vary=False)
        self.params.add('k0', value=0.39, min=0, max=np.inf, vary=False)
        self.params.add('kq', value=1.34e4, min=0, max=np.inf, vary=False)
        self.params.add('kd', value=0.186, min=0, max=np.inf, vary=False)
        self.params.add('kTT', value=1e4, min=0, max=np.inf, vary=False)
        self.params.add('kdel', value=1e4, min=0, max=np.inf, vary=True)
        self.params.add('cCP1', value=1e-4, min=0, max=np.inf, vary=False)
        self.params.add('y0', value=0, min=-np.inf, max=np.inf, vary=False)

    def initialize_values(self, x_data, y_data):
        # self.params['A0'].value = np.max(y_data[0])
        # self.params['B'].value = y_data[0]
        pass

    @staticmethod
    def func(x, Xan, DR, DF, A0, kex, k_ex, k0, kq, kd, kTT, kdel, cCP1, y0):
        def Tall(A):
            result = (1 - np.power(10, -A)) / (A * np.log(10))
            result[np.isnan(result)] = 1
            return result

        def solve(profiles, t, *args):
            Xan, Ex, DR, DF = profiles
            k1_var = args[0]

            dcXan_dt = -(k0 + k1_var) * Xan
            dcEx_dt = + k1_var * Xan - (kex + kd) * Ex + k_ex * DR
            dcDR_dt = + kex * Ex - kd * DR - k_ex * DR
            # DF simulation
            dcDF_dt = kTT * Xan * DR - kdel * DF

            return [dcXan_dt, dcEx_dt, dcDR_dt, dcDF_dt]

        x0 = np.linspace(0, x[0], num=100)
        _init_x = odeint(solve, [1, 0, 0, 0], x0, args=(kq * cCP1,))[-1, :]  # take the row in the result matrix

        result = odeint(solve, _init_x, x, args=(kq * cCP1,))
        cXan = result[:, 0]
        cEx = result[:, 1]
        cDR = result[:, 2]
        cDF = result[:, 3] * (Tall(A0 * cXan) if A0 > 0 else 1)

        return Xan * cXan + DR * (cEx + cDR) + DF * cDF / np.max(cDF) + y0


class DF_max(_Model):
    """l"""

    name = 'DF Maxima'

    def init_params(self):
        self.params.add('A0', value=0, min=0, max=10, vary=False)
        self.params.add('kex', value=1.5, min=0, max=np.inf, vary=True)
        self.params.add('k_ex', value=0, min=0, max=np.inf, vary=False)
        self.params.add('k0', value=0.128296, min=0, max=np.inf, vary=False)
        self.params.add('kq', value=13440.21, min=0, max=np.inf, vary=False)
        self.params.add('kd', value=0.19101218, min=0, max=np.inf, vary=False)
        self.params.add('kTT', value=1e4, min=0, max=np.inf, vary=False)
        self.params.add('kdel', value=1e4, min=0, max=np.inf, vary=True)
        self.params.add('ksq', value=0, min=0, max=np.inf, vary=True)
        self.params.add('points', value=500, min=1, max=np.inf, vary=False)
        self.params.add('t_lim', value=6, min=0, max=np.inf, vary=False)
        self.params.add('T', value=20, min=-30, max=50, vary=False)

    def initialize_values(self, x_data, y_data):
        # self.params['A0'].value = np.max(y_data[0])
        # self.params['B'].value = y_data[0]
        pass

    @staticmethod
    def func(x, A0, kex, k_ex, k0, kq, kd, kTT, kdel, ksq, points, t_lim, T):
        intensity = False

        # from fit_script import get_constants
        #
        # T_index = int((T + 30) * 8 / 80)
        # k0, kq, kd = get_constants(T=T, T_index=T_index)

        def Tall(A):
            result = (1 - np.power(10, -A)) / (A * np.log(10))
            result[np.isnan(result)] = 1
            return result

        def solve(profiles, t, *args):
            Xan, Ex, DR, DF = profiles
            cCP1 = args[0]

            dcXan_dt = -(k0 + kq * cCP1 + ksq * cCP1) * Xan
            dcEx_dt = + kq * cCP1 * Xan - (kex + kd) * Ex + k_ex * DR
            dcDR_dt = + kex * Ex - kd * DR - k_ex * DR
            # DF simulation
            dcDF_dt = kTT * Xan * DR - kdel * DF

            return [dcXan_dt, dcEx_dt, dcDR_dt, dcDF_dt]

        def integrate(k1, times, init_values):
            result = odeint(solve, init_values, times, args=(k1,))
            Xan = result[:, 0]
            Ex = result[:, 1]
            DR = result[:, 2]
            DF = result[:, 3]
            return Xan, Ex, DR, DF

        def max(k1):
            limit_t = t_lim
            num_p = points
            t = np.linspace(0, limit_t, num_p)

            Xan, Ex, DR, DF = integrate(k1, t, [1, 0, 0, 0])
            DF_corr = DF * (Tall(A0 * Xan) if A0 > 0 else 1)
            t_max_idx = np.argmax(DF_corr)

            min_idx = t_max_idx - 1 if t_max_idx > 0 else 0
            t_part = np.linspace(t[min_idx], t[t_max_idx + 1], num_p)
            Xan, Ex, DR, DF = integrate(k1, t_part, [Xan[min_idx], Ex[min_idx], DR[min_idx], DF[min_idx]])

            DF_corr = DF * (Tall(A0 * Xan) if A0 > 0 else 1)

            return t_part[np.argmax(DF_corr)] if not intensity else np.max(DF_corr)

        maxima = np.zeros(x.shape[0])

        for i, k1 in enumerate(x):
            maxima[i] = max(k1)

        return maxima


#
# class DelayedFluorescence2(_Model):
#     """l"""
#
#     name = 'Delayed Fluorescence2'
#
#     def init_params(self):
#         # self.params.add('c0', value=1, min=0, max=np.inf, vary=False)
#         self.params.add('Xan', value=0, min=-np.inf, max=np.inf, vary=False)
#         self.params.add('DR', value=0, min=-np.inf, max=np.inf, vary=False)
#         self.params.add('DF', value=1, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('A0', value=0.2, min=-np.inf, max=np.inf, vary=False)
#
#         # self.params.add('B', value=1, min=-np.inf, max=np.inf, vary=True)
#         # self.params.add('C', value=0, min=-np.inf, max=np.inf, vary=False)
#         self.params.add('k0', value=0.4, min=0, max=np.inf, vary=False)
#         self.params.add('k1', value=1, min=0, max=np.inf, vary=False)
#         self.params.add('kd', value=0.186, min=0, max=np.inf, vary=False)
#         self.params.add('kex', value=0.5, min=0, max=np.inf, vary=True)
#
#         # self.params.add('k3', value=1, min=0, max=np.inf, vary=True)
#         # self.params.add('k4', value=1, min=0, max=np.inf, vary=True)
#
#         self.params.add('y0', value=0, min=-np.inf, max=np.inf, vary=True)
#
#     def initialize_values(self, x_data, y_data):
#         # self.params['A'].value = np.max(y_data[0])
#         # self.params['B'].value = y_data[0]
#         pass
#
#     # @nb.njit(fastmath=True)
#     # @staticmethod
#     # def sum(n=1000, A0=0):
#     #     sum = 0
#     #     for i in range(0, n):
#     #         sum += np.exp(-i / (n - 1) * A0)
#
#     @staticmethod
#     def func(x, Xan, DR, DF, A0, k0, k1, kd, kex, y0):
#
#         def Tall(A, classical=False):
#             if classical:
#                 return np.power(10, -A)
#             else:
#                 result = (1 - np.power(10, -A)) / (A * np.log(10))
#                 result[np.isnan(result)] = 1
#                 return result
#
#         def solve(conc, t):
#             cA, cB, cC = conc
#
#             dcA_dt = -(k0 + k1) * cA
#             dcB_dt = + k1 * cA - kex * cB - kd * cB
#             dcC_dt = + kex * cB - kd * cC
#
#             return [dcA_dt, dcB_dt, dcC_dt]
#
#         x0 = np.linspace(0, x[0], num=100)
#         _init_x = odeint(solve, [1, 0, 0], x0)[-1, :]  # take the row in the result matrix
#
#         result = odeint(solve, _init_x, x)
#         cXan = result[:, 0]
#         cEx = result[:, 1]
#         cDR = result[:, 2]
#         cDF = cXan * cDR
#
#         cDF *= Tall(cXan * A0) if A0 > 0 else 1  # selfabsorption correction
#
#         return Xan * cXan + DR * (cEx + cDR) + DF * cDF / np.max(cDF) + y0


class ABC1st_XanModel(_Model):
    """This is the A→B→C model, 1st order kinetics. Differential equations:

    :math:`\\frac{\mathrm{d}[\mathrm{A}]}{\mathrm{d}t} = -k_1[\mathrm{A}]`

    :math:`\\frac{\mathrm{d}[\mathrm{B}]}{\mathrm{d}t} = +k_1[\mathrm{A}] - k_2[\mathrm{B}]`

    :math:`\\frac{\mathrm{d}[\mathrm{C}]}{\mathrm{d}t} = +k_2[\mathrm{B}]`

    initial conditions: :math:`[\mathrm A]_0 = c_0 \qquad [\mathrm B]_0 = [\mathrm C]_0 = 0`
    """

    name = 'A→B→C (1st order-Xan-k0)'

    def init_params(self):
        self.params.add('A', value=1, min=-np.inf, max=np.inf, vary=True)
        self.params.add('B', value=1, min=-np.inf, max=np.inf, vary=True)
        self.params.add('C', value=0, min=-np.inf, max=np.inf, vary=False)
        self.params.add('k0', value=0.4, min=0, max=np.inf, vary=False)
        self.params.add('k1', value=1, min=0, max=np.inf, vary=True)
        self.params.add('k2', value=0.5, min=0, max=np.inf, vary=True)
        self.params.add('y0', value=0, min=-np.inf, max=np.inf, vary=True)

    def initialize_values(self, x_data, y_data):
        self.params['A'].value = y_data[0]
        self.params['B'].value = y_data[0]

    @staticmethod
    def func(x, A, B, C, k0, k1, k2, y0):
        cA = np.exp(-(k0 + k1) * x)

        # if k1 == k2, the integrated law fow cB has to be changed
        if np.abs(k1 - k2) < 1e-10:
            cB = (k1 * (np.exp(k0 * x) - 1) * np.exp(x * (-(k0 + k1)))) / k0
        else:  # we would get zero division error for k1 = k2
            cB = (k1 * (np.exp(-k2 * x) - np.exp(-x * (k0 + k1)))) / (k0 + k1 - k2)

        cC = 1 - cA - cB

        return A * cA + B * cB + C * cC + y0


class Equlibrium(_Model):
    """This is the A→B→C model, 1st order kinetics. Differential equations:

    :math:`\\frac{\mathrm{d}[\mathrm{A}]}{\mathrm{d}t} = -k_1[\mathrm{A}]`

    :math:`\\frac{\mathrm{d}[\mathrm{B}]}{\mathrm{d}t} = +k_1[\mathrm{A}] - k_2[\mathrm{B}]`

    :math:`\\frac{\mathrm{d}[\mathrm{C}]}{\mathrm{d}t} = +k_2[\mathrm{B}]`

    initial conditions: :math:`[\mathrm A]_0 = c_0 \qquad [\mathrm B]_0 = [\mathrm C]_0 = 0`
    """

    name = 'Equilibrium model A+2L=2B, A+L=B and A+2L=B'

    def init_params(self):
        self.params.add('c0', value=1, min=0, max=np.inf, vary=True)
        self.params.add('epsA', value=1000, min=0, max=np.inf, vary=True)
        self.params.add('epsB', value=1000, min=0, max=np.inf, vary=True)
        self.params.add('K', value=2, min=0, max=np.inf, vary=True)
        self.params.add('model', value=0, min=0, max=2, vary=False)

    # def initialize_values(self, x_data, y_data):
    #     self.params['A'].value = y_data[0]
    #     self.params['B'].value = y_data[0]

    @staticmethod
    def func(x, c0, epsA, epsB, K, model):
        cL = x

        if model == 0:
            alpha = cL * (-K * cL + np.sqrt(K * (K * cL * cL + 16 * c0))) / (8 * c0)

            # alpha = np.zeros(cL.shape[0])
            # for i, cLi in enumerate(cL):
            #     polyi = [4*K*c0*c0, -4*K*c0*c0 - 4*K*c0*cLi + 4*c0, 4*K*c0*cLi + K*cLi*cLi, -K*cLi*cLi]
            #     roots = np.roots(polyi)  # find roots of this polynomial
            #     alphai = roots[roots >= 0][0]  # take first positive or zero root
            #
            #     alpha[i] = alphai

        elif model == 1:
            alpha = K * cL / (K * cL + 1)
        else:
            alpha = cL * (-K * cL + np.sqrt(K * (K * cL * cL + 4 * c0))) / (2 * c0)

            # alpha = K * cL * cL / (K * cL * cL + 1)

        return c0 * (epsA + alpha * ((2 if model == 0 else 1) * epsB - epsA))


if __name__ == '__main__':
    # m = LinearModel()

    import matplotlib.pyplot as plt

    # print(m.wrapper_func(np.asarray([0, 0.1, 0.15, 0.3, 0.6]), m.params))

    x = np.linspace(0, 10, 1000)

    m = DelayedFluorescence2()

    m.params['A'].value = 1
    m.params['A0cA'].value = 1
    m.params['k0'].value = 0.3

    m.params['k1'].value = 1.95
    m.params['kd'].value = 0.2

    n = DelayedFluorescence2()

    n.params['A'].value = 1
    n.params['A0cA'].value = 0
    n.params['k0'].value = 0.3
    n.params['k1'].value = 1.95
    n.params['kd'].value = 0.2

    mr = m.wrapper_func(x, m.params)
    nr = n.wrapper_func(x, n.params)

    plt.plot(x, mr, label='DelayedFluorescence2 A0 1')
    plt.plot(x, nr, label='DelayedFluorescence2 A0 0')

    plt.legend()
    plt.show()



from PyQt5 import QtCore


class Task(QtCore.QThread):  # https://doc.qt.io/qt-5/qthread.html

    def __init__(self, parent=None):
        super(Task, self).__init__(parent)
        self.started.connect(self.preRun)
        self.finished.connect(self.postRun)

    def preRun(self):
        """
        The code in this method is runs before run in GUI thread.
        """
        pass

    def run(self):
        """
        The code in this method is run in another thread.
        """
        pass

    def postRun(self):
        """
        The code in this method is run in GUI thread.
        """
        pass

